#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile
import json

import pytest

from poppy.core.logger import logger
from poppy.core.test import CommandTestCase

# Define temporary folder for test
from roc.idb.models.idb import PacketHeader, ItemInfo, IdbRelease

TEST_TEMP_DIR = tempfile.gettempdir()


class TestPacketsToJson(CommandTestCase):
    """
    Test the rpl pkt_to_json command.
    """

    @pytest.mark.parametrize(
        'idb_source,idb_version,output_json_file',
        [('MIB', '20200131',
          os.path.join(TEST_TEMP_DIR, 'test_packets_to_json.json'))],
    )
    def test_packets_to_json(self, idb_source, idb_version, output_json_file):
        from poppy.core.conf import Settings

        # initialize the IDB

        # apply database migrations
        db_upgrade = ['pop', 'db', 'upgrade', 'heads', '-ll', 'INFO']

        # IDB PALISADE loading
        idb_palisade_loading = [
            'pop',
            'idb',
            'install',
            '-i',
            os.environ.get('IDB_INSTALL_DIR', tempfile.gettempdir()),
            '-s',
            'PALISADE',
            '-v',
            '4.3.5_MEB_PFM',
            '--load',
            '-ll',
            'INFO',
        ]

        # IDB MIB loading
        idb_mib_loading = [
            'pop',
            'idb',
            'install',
            '-i',
            os.environ.get('IDB_INSTALL_DIR', tempfile.gettempdir()),
            '-s',
            idb_source,
            '-v',
            idb_version,
            '--load',
            '--current',
            '-ll',
            'INFO',
        ]

        # apply database migrations
        self.run_command(db_upgrade)

        # Apply IDB loading
        self.run_command(idb_palisade_loading)

        # Apply IDB loading
        self.run_command(idb_mib_loading)

        # Ensure IDB is loaded to continue
        if self.is_idb(idb_source, idb_version):

            # Build command to test
            binary = '0CB7D904000D100501002730B9A4E54FA6D20165'
            command_to_test = ['pop', 'rpl',
                            '--idb-version', idb_version,
                            '--idb-source', idb_source,
                            'pkt_to_json',
                            '--binary', binary,
                            '--output-json-file', output_json_file,
                            '-ll', 'WARNING']
            self.run_command(command_to_test)

            assert os.path.isfile(output_json_file)

            # Open expected output and test output JSON files and compare both
            test_data_path = os.path.abspath(os.path.dirname(__file__))
            expected_json_path = os.path.join(test_data_path , 'data', os.path.basename(output_json_file))
            with open(expected_json_path, 'r') as ejson:
                exp_json = json.load(ejson)

            with open(output_json_file, 'r') as ojson:
                out_json = json.load(ojson)

            assert exp_json == out_json

        else:
            logger.error(f'Loading IDB [{idb_source}-{idb_version} has failed!')
            assert False


    @pytest.mark.parametrize(
        'idb_source,idb_version,output_json_file',
        [('MIB', '20200131',
          os.path.join(TEST_TEMP_DIR, 'test_packets_to_json_header_only.json'))],
    )
    def test_packets_to_json_header_only(self, idb_source, idb_version, output_json_file):
        from poppy.core.conf import Settings

        # initialize the IDB

        # Db migration
        db_upgrade = ['pop', 'db', 'upgrade', 'heads', '-ll', 'INFO']

        # IDB PALISADE loading
        idb_palisade_loading = [
            'pop',
            'idb',
            'install',
            '-i',
            os.environ.get('IDB_INSTALL_DIR', tempfile.gettempdir()),
            '-s',
            'PALISADE',
            '-v',
            '4.3.5_MEB_PFM',
            '--load',
            '-ll',
            'INFO',
        ]

        # IDB MIB loading
        idb_mib_loading = [
            'pop',
            'idb',
            'install',
            '-i',
            os.environ.get('IDB_INSTALL_DIR', tempfile.gettempdir()),
            '-s',
            idb_source,
            '-v',
            idb_version,
            '--load',
            '--current',
            '-ll',
            'INFO',
        ]

        # apply database migrations
        self.run_command(db_upgrade)

        # Apply IDB loading
        self.run_command(idb_palisade_loading)

        # Apply IDB loading
        self.run_command(idb_mib_loading)

        # Ensure IDB is loaded to continue
        if self.is_idb(idb_source, idb_version):

            # Build command to test
            binary = '0CB7D904000D100501002730B9A4E54FA6D20165'
            command_to_test = ['pop', 'rpl',
                            '--idb-version', idb_version,
                            '--idb-source', idb_source,
                            'pkt_to_json',
                            '--binary', binary,
                            '--output-json-file', output_json_file,
                            '-ll', 'ERROR']
            self.run_command(command_to_test)

            assert os.path.isfile(output_json_file)

            # Open expected output and test output JSON files and compare both
            test_data_path = os.path.abspath(os.path.dirname(__file__))
            expected_json_path = os.path.join(test_data_path , 'data', os.path.basename(output_json_file))
            with open(expected_json_path, 'r') as ejson:
                exp_json = json.load(ejson)

            with open(output_json_file, 'r') as ojson:
                out_json = json.load(ojson)

            assert exp_json == out_json

        else:
            logger.error(f'Loading IDB [{idb_source}-{idb_version} has failed!')
            assert False

    def is_idb(self, idb_source, idb_version):

        logger.debug(f'Querying IDB [{idb_source}-{idb_version}] ...')
        try:
            packet_header = (
                self.session.query(PacketHeader)
                .join(ItemInfo)
                .join(IdbRelease)
                .filter(ItemInfo.srdb_id == 'YIW00083',
                        IdbRelease.idb_version == idb_version,
                        IdbRelease.idb_source == idb_source)
                .one()
            )
        except:
            return False
        else:
            return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path

from poppy.core.generic.paths import Paths

__all__ = ["paths"]

_ROOT_DIRECTORY = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
    )
)

paths = Paths(_ROOT_DIRECTORY)




from libc.stdint cimport uint8_t
from libc.stdint cimport uint32_t
from libc.stdint cimport int32_t

cdef extern from "GsrPreprocessMode.h":
    ctypedef enum PreprocessMode:
        ESTIMATE_1D_H

cdef extern from "GsrRiceUncompress.h":
    # function to uncompress the data
    int uncompress(unsigned char* compressed, int compressedLen, unsigned int* data, int nbInput) except +

    # post process data
    int postprocessor(unsigned int* data, int nbInput, unsigned short* postProcessed, PreprocessMode mode) except +

cdef extern from "GsrRiceCompress.h":
    # function to compress the data
    int compress(unsigned int* data, int nbInput, unsigned char* compressed, int compressedLen) except +

    # preprocess the data
    int preprocess(unsigned short* data, int nbInput, unsigned int* preprocessed, PreprocessMode mode) except +



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from sqlalchemy import func

try:
    from poppy.core.db.connector import Connector
    from poppy.core.logger import logger
except:
    print('POPPy framework is not installed')

try:
    from roc.idb.models.idb import PacketHeader, ItemInfo
except:
    print('roc.idb plugin is not installed!')

from roc.rpl.exceptions import InvalidPacketID
from roc.rpl.packet_parser.packet_cache import PacketCache
from roc.rpl.packet_structure.data import Data
from roc.rpl.packet_structure.data import header_foot_offset


__all__ = ['ExtendedData']


class ExtendedData(Data):
    """
    Extended data class which allows to manage the analysis of the packet header
    as well as the identification of the PALISADE ID.
    """

    # Use __new__ instead of __init__ since Data parent class is supposed as immutable class
    # __new__ is then required to introduce idb ExtendedData class argument
    # FIXME - See why roc.rap installation fails when using settings.PIPELINE_DATABASE
    #@Connector.if_connected('MAIN-DB')
    def __new__(cls, data, size, idb):

        instance = super(ExtendedData, cls).__new__(cls, data, size)

        # Get/create PacketCache singleton instance
        instance._cache = PacketCache()

        # Initialize IDB
        if idb in instance._cache.idb:
            instance.idb = instance._cache.idb[idb][0]
            instance.session = instance._cache.idb[idb][1]
        else:
            instance.idb = idb.get_idb()
            instance.session = idb.session
            instance._cache.idb[idb] = (instance.idb, instance.session)

        return instance

    def _get_max_sid(self, header, data_header):
        """
        Get Packet max_sid value

        :param header: packet_header parameters
        :param data_header: packet data_field_header parameters
        :return: max_sid value
        """

        # Return Sid value if already in the cache
        param_tuple = (self.idb,
                       header.packet_type,
                       header.process_id,
                       header.packet_category,
                       data_header.service_type,
                       data_header.service_subtype)
        if param_tuple in self._cache.max_sid:
            return self._cache.max_sid[param_tuple]

        # Else, retrieve value from IDB data in the database

        # The structure_id (sid) is required to fully identify TM packets
        if header.packet_type == 0:
            query_max_sid = self.session.query(func.max(PacketHeader.sid))
            query_max_sid = query_max_sid.join(ItemInfo)
            query_max_sid = query_max_sid.filter_by(idb=self.idb)

            query_max_sid = query_max_sid.filter(
                PacketHeader.cat == header.packet_category)
            query_max_sid = query_max_sid.filter(
                PacketHeader.pid == header.process_id)
            if data_header is not None:
                query_max_sid = query_max_sid.filter(
                    PacketHeader.service_type == data_header.service_type)
                query_max_sid = query_max_sid.filter(
                    PacketHeader.service_subtype == data_header.service_subtype)

            max_sid_value = query_max_sid.one()[0]
            # now we can obtain the palisade name
        else:
            max_sid_value = None

        # add max_sid_value in cache
        self._cache.max_sid[param_tuple] = max_sid_value

        return max_sid_value

    def _get_sid(self, header, data_header, max_sid_value):
        """
        Get Packet sid value

        :param header: packet_header parameters
        :param data_header: packet data_field_header parameters
        :param max_sid_value: max sid value
        :return: sid value
        """

        # get header and footer offsets
        header_offset, _ = header_foot_offset(header.packet_type)

        # For service 1 TM packets, add 4 bytes to the offset
        if data_header.service_type == 1:
            header_offset += 4

        # get the value from data
        if max_sid_value <= 255:
            value = self.u8p(header_offset, 0, 8)
            # logger.warning('8 bits data')

        else:
            value = self.u16p(header_offset, 0, 16)
            # logger.warning('16 bits data')

        return value

    def _get_packet_ids(self, header, data_header):
        """
        Get packet IDs (SRDB_ID and PALISADE ID)

        :param header: packet_header parameters
        :param data_header: packet data_field_header parameters
        :param max_sid_value: max SID value
        :param sid_value:
        :return: tuple (SRDB ID, PALISADE ID) for the current packet
        """

        # Initialize outputs
        srdb_id = None
        palisade_id = None

        # extract the Structure ID (SID) only if the packet is not empty ...

        # get the max possible value for the SID
        max_sid_value = self._get_max_sid(header, data_header)

        # If it is a TM packet and it has a SID, then...
        if max_sid_value is not None:
            # SID is always positive, so if max_sid_value == 0, then ...
            if max_sid_value == 0:
                sid_value = 0
            else:
                # check the first byte of data to get the SID
                sid_value = self._get_sid(header, data_header, max_sid_value)
        else:
            sid_value = None

        # Return SRDB ID and palisade ID if already stored in the cache
        param_tuple = (self.idb,
                       header.packet_category,
                       header.process_id,
                       data_header.service_type,
                       data_header.service_subtype,
                       max_sid_value, sid_value)

        if param_tuple in self._cache.packet_ids:
            return self._cache.packet_ids[param_tuple]

        # Else, build query to retrieve the TM/TC packet srdb and palisade IDs
        query_packet_ids = self.session.query(
            ItemInfo.srdb_id, ItemInfo.palisade_id)
        query_packet_ids = query_packet_ids.filter_by(idb=self.idb)
        query_packet_ids = query_packet_ids.join(PacketHeader)
        query_packet_ids = query_packet_ids.filter(
            PacketHeader.cat == header.packet_category)
        query_packet_ids = query_packet_ids.filter(
            PacketHeader.pid == header.process_id)
        # logger.warning('packet_category: '+str(header.packet_category))
        # logger.warning('process_id: '+str(header.process_id))
        if data_header is not None:
            query_packet_ids = query_packet_ids.filter(
                PacketHeader.service_type == data_header.service_type)
            query_packet_ids = query_packet_ids.filter(
                PacketHeader.service_subtype == data_header.service_subtype)
        # logger.warning('max sid value: '+str(max_sid_value))

        # If it is a TM packet and it has a SID, then...
        if max_sid_value is not None:
            # SID is always positive, so if max_sid_value == 0, then ...
            if max_sid_value == 0:
                sid_value = 0
            else:
                sid_value = self._get_sid(header, data_header, max_sid_value)
            query_packet_ids = query_packet_ids.filter(
                PacketHeader.sid == sid_value)

        try:
            srdb_id, palisade_id = query_packet_ids.all()[0]
        except InvalidPacketID:
            logger.warning('Unknown palisade ID')
        except:
            logger.exception('Querying packet IDs has failed!')
        else:
            # add into the cache
            self._cache.packet_ids[param_tuple] = (srdb_id, palisade_id)
        finally:
            return (srdb_id, palisade_id)

    def get_packet_summary(self,
                           srdb_id=None,
                           palisade_id=None):
        """
        Get a summary (srdb_id, palisade_id, header, data_header) of the packet.

        :return: srdb_id, palisade_id, header, data_header
        """

        # extract the header
        header = self.extract_header()

        # check if there is a data header
        data_header = None
        if header.data_field_header_flag:
            # check if the packet is a TM or a TC
            packet_type = header.packet_type
            if packet_type == 0:
                data_header = self.extract_tm_header()
            elif packet_type == 1:
                data_header = self.extract_tc_header()

            # get srdb and palisade ids
            if srdb_id is None or palisade_id is None:
                srdb_id, palisade_id = self._get_packet_ids(
                    header, data_header)

        else:
            logger.warning('No data header')

        return srdb_id, palisade_id, header, data_header



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path as osp
import struct
import unittest

from roc.rpl.packet_structure.data import Data
from roc.rpl.rice.rice import Compressor


class TestRICE(unittest.TestCase):
    """
    To test the algorithm of the compression and decompression of data.
    """

    def setUp(self):
        """
        Create the list of data to compress and decompress for tests.
        """
        self.data = [
            [
                0, 0, 0, 0,
                0, 0, 0, 0,
                0, 0, 0, 0,
                0, 0, 0, 0,
            ],
            [
                2, 4, 89, 233,
                3, 34, 4, 8,
                2, 4, 89, 233,
                3, 34, 4, 8,
            ],
            [
                1, 0, 0, 0,
                0, 1, 2, 9,
                1, 0, 0, 0,
                0, 1, 2, 9,
            ],
            [
                0, 0, 0, 1,
                0, 2, 0, 3,
                0, 4, 0, 5,
                0, 6, 0, 7,
                0, 8, 0, 9,
                0, 10, 0, 11,
                0, 12, 0, 13,
                0, 14, 0, 15,
            ],
        ]

        # open the file with data for test
        path = osp.abspath(osp.dirname(__file__))
        path = osp.join(path, 'data', 'data.dat')
        with open(path, 'r') as f:
            # read a line corresponding to a packet
            for line in f.readlines():
                # append the data
                data = [int(x) for x in line.split()]
                self.data.append(data)

        # convert into unsigned int bytes array
        self.byte_data = [
            bytes(struct.pack('>' + str(len(x)) + 'h', *x))
            for x in self.data
        ]

    def test_compress(self):
        """
        Just to check compression is good.
        """
        # run the test more than once to track segfault if possible
        for i in range(20):
            # loop over data
            for buf in self.byte_data:
                #  print("Buffer", buf, len(buf))
                self.compressor = Compressor()
                self.compressed = self.compressor.compress(
                    Data(
                        buf,
                        len(buf),
                    )
                )
                self.compressor = Compressor()
                #  print("Compressed", self.compressed, self.csize)
                self.uncompressed = self.compressor.uncompress(
                    self.compressed,
                    len(buf),
                )
                #  print("Uncompressed", self.uncompressed[0])

                # the comparison is done for the given size of the data and not
                # the entire data because there is a minimal size for the data
                # in output if the decompression, because they work by block
                #  print("Number", i)
                self.assertEqual(
                    buf[:len(buf)],
                    self.uncompressed.to_bytes()[:len(buf)],
                )
                #  print("")
                #  print("#################")
                #  print("\n"*3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Exceptions definition for PacketParser plugin.
"""

try:
    from poppy.core.logger import logger
except:
    print('POPPy framework is not installed!')

__all__ = ['RPLError', 'InvalidPacketID',
            'SpiceError','SpiceKernelNotValid', 'NoSpiceFound',
           'TransferFunctionError', 'PacketParsingError']


def init(self, class_obj, message, *args, **kwargs):
    super(class_obj, self).__init__(*args, **kwargs)
    logger.error(message)
    self.message = message

class InvalidPacketID(Exception):
    """
    Exception raised when a packet has an invalid packet
    """
    def __init__(self, message, *args, **kwargs):
        init(self, InvalidPacketID, message, *args, **kwargs)

class RPLError(Exception):
    """
    Exception raised when an error has occurred with RPL
    """
    def __init__(self, message, *args, **kwargs):
        init(self, RPLError, message, *args, **kwargs)

class SpiceError(Exception):
    """
    Exception raised when an error has occurred with SPICE
    """
    def __init__(self, message, *args, **kwargs):
        init(self, SpiceError, message, *args, **kwargs)

class NoSpiceFound(Exception):
    """
    Exception SPICE not loaded
    """
    pass

class SpiceKernelNotValid(Exception):
    """
    Exception for badly formatted Spice kernel
    """
    pass


class PacketParsingError(Exception):
    """
    Exception raised when an error has occurred when parsing packets
    """
    def __init__(self, message, *args, **kwargs):
        init(self, PacketParsingError, message, *args, **kwargs)

class TransferFunctionError(Exception):
    """
    Exception raised when TransferFunction class call has failed
    """
    def __init__(self, message, *args, **kwargs):
        init(self, TransferFunctionError, message, *args, **kwargs)



cdef class Header:
    def __cinit__(
        Header self,
        uint8_t ccsds_version_number,
        uint8_t packet_type,
        uint8_t data_field_header_flag,
        uint8_t process_id,
        uint8_t packet_category,
        uint8_t segmentation_grouping_flag,
        uint16_t sequence_cnt,
        uint16_t packet_length,
    ):
        self.ccsds_version_number = ccsds_version_number
        self.packet_type = packet_type
        self.data_field_header_flag = data_field_header_flag
        self.process_id = process_id
        self.packet_category = packet_category
        self.segmentation_grouping_flag = segmentation_grouping_flag
        self.sequence_cnt = sequence_cnt
        self.packet_length = packet_length

    cpdef tuple to_tuple(Header self):
        """
        Return a tuple of the values of the header.
        """
        return (
            self.ccsds_version_number,
            self.packet_type,
            self.data_field_header_flag,
            self.process_id,
            self.packet_category,
            self.segmentation_grouping_flag,
            self.sequence_cnt,
            self.packet_length,
        )

    cpdef dict to_dict(Header self):
        """
        Return a dictionary of the values of the header.
        """
        return {
            'ccsds_version_number': self.ccsds_version_number,
            'packet_type': self.packet_type,
            'data_field_header_flag': self.data_field_header_flag,
            'process_id': self.process_id,
            'packet_category': self.packet_category,
            'segmentation_grouping_flag': self.segmentation_grouping_flag,
            'sequence_cnt': self.sequence_cnt,
            'packet_lenght': self.packet_length,
        }


    def __repr__(Header self):
        return (
            "Header(" +
            "ccsds_version_number={0}, ".format(self.ccsds_version_number) +
            "packet_type={0}, ".format(self.packet_type) +
            "data_field_header_flag={}, ".format(self.data_field_header_flag) +
            "process_id={0}, ".format(self.process_id) +
            "packet_category={0}, ".format(self.packet_category) +
            "segmentation_grouping_flag={0}, ".format(
                self.segmentation_grouping_flag
            ) +
            "sequence_cnt={0}, ".format(self.sequence_cnt) +
            "packet_length={0})".format(self.packet_length)
        )


cdef class TCHeader:
    def __cinit__(
        TCHeader self,
        uint8_t ccsds_secondary_header_flag,
        uint8_t pus_version,
        uint8_t ack_execution_completion,
        uint8_t ack_execution_progress,
        uint8_t ack_execution_start,
        uint8_t ack_acceptance,
        uint8_t service_type,
        uint8_t service_subtype,
        uint8_t source_id,
    ):
        self.ccsds_secondary_header_flag = ccsds_secondary_header_flag
        self.pus_version = pus_version
        self.ack_execution_completion = ack_execution_completion
        self.ack_execution_progress = ack_execution_progress
        self.ack_execution_start = ack_execution_start
        self.ack_acceptance = ack_acceptance
        self.service_type = service_type
        self.service_subtype = service_subtype
        self.source_id = source_id

    cpdef tuple to_tuple(TCHeader self):
        """
        Return the header with tuple.
        """
        return (
            self.ccsds_secondary_header_flag,
            self.pus_version,
            self.ack_execution_completion,
            self.ack_execution_progress,
            self.ack_execution_start,
            self.ack_acceptance,
            self.service_type,
            self.service_subtype,
            self.source_id,
        )

    cpdef dict to_dict(TCHeader self):
        """
        Return the header with dictionary.
        """
        return {
            'ccds_secondary_header_flag': self.ccsds_secondary_header_flag,
            'pus_version': self.pus_version,
            'ack_execution_completion': self.ack_execution_completion,
            'ack_execution_progress': self.ack_execution_progress,
            'ack_execution_start': self.ack_execution_start,
            'ack_acceptance': self.ack_acceptance,
            'service_type': self.service_type,
            'service_subtype': self.service_subtype,
            'source_id': self.source_id,
        }

    def __repr__(TCHeader self):
        return (
            "TCHeader(" +
            "ccsds_secondary_header_flag={0},".format(
                self.ccsds_secondary_header_flag
            ) +
            "pus_version={0},".format(self.pus_version) +
            "ack_execution_completion={0},".format(
                self.ack_execution_completion
            ) +
            "ack_execution_progress={0},".format(self.ack_execution_progress) +
            "ack_execution_start={0},".format(self.ack_execution_start) +
            "ack_acceptance={0},".format(self.ack_acceptance) +
            "service_type={0},".format(self.service_type) +
            "service_subtype={0},".format(self.service_subtype) +
            "source_id={0}".format(self.source_id) +
            ")"
        )


cdef class TMHeader:
    def __cinit__(
        TMHeader self,
        uint8_t spare_1,
        uint8_t pus_version,
        uint8_t spare_2,
        uint8_t service_type,
        uint8_t service_subtype,
        uint8_t destination_id,
        tuple time,
    ):
        self.spare_1 = spare_1
        self.pus_version = pus_version
        self.spare_2 = spare_2
        self.service_type = service_type
        self.service_subtype = service_subtype
        self.destination_id = destination_id
        self.time = time

    cpdef tuple to_tuple(TMHeader self):
        return (
            self.spare_1,
            self.pus_version,
            self.spare_2,
            self.service_type,
            self.service_subtype,
            self.destination_id,
            self.time,
        )

    cpdef dict to_dict(TMHeader self):
        return {
            'spare_1': self.spare_1,
            'pus_version': self.pus_version,
            'spare_2': self.spare_2,
            'service_type': self.service_type,
            'service_subtype': self.service_subtype,
            'destination_id': self.destination_id,
            'time': self.time,
        }

    def __repr__(TMHeader self):
        return (
            "TMHeader(" +
            "spare_1={0},".format(self.spare_1) +
            "pus_version={0},".format(self.pus_version) +
            "spare_2={0},".format(self.spare_2) +
            "service_type={0},".format(self.service_type) +
            "service_subtype={0},".format(self.service_subtype) +
            "destination_id={0},".format(self.destination_id) +
            "time={0}".format(self.time) +
            ")"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import struct

from roc.rpl.exceptions import RPLError
from roc.rpl.packet_parser.packet_parser import PacketParser
from .utils import find_index
from .utils import get_base_packet
from roc.rpl.packet_structure.data import Data

__all__ = []


def get_base_tds(Packet):
    """
    Generate the base class to use for compressed packets in the case of TDS.
    """
    # base class for compressed packet
    Base = get_base_packet(Packet)

    # base class for the compressed packets of LFR
    class TDS(Base):
        """
        Base class for the compressed packets of LFR. Define the special
        behaviour of the treatment for those packets, that should mimic the non
        compressed (NC) ones.
        """
        def __init__(self, name):
            # init the packet as the other
            super(TDS, self).__init__(name)

            # the index of the parameter with the number of blocks
            self.idx_nchannels, self.nchannels = find_index(
                self.parameters,
                self.NCHANNEL,
            )

        def data_substitution(self, data, uncompressed):
            """
            Create a new data bytes array in order to simulate an non
            compressed packet with the data from an uncompressed packet,
            allowing to decommute and store a compressed packet as an
            uncompressed one. Different from the base one since we need to add
            the parameter indicating the number of blocks used in the packet,
            not present in the compressed packet.
            """
            new_bytes = data.to_bytes()[:self.start.byte + self.header_offset]
            new_bytes += struct.pack(
                ">H",
                len(uncompressed) // self.block_size,
            )
            new_bytes += uncompressed
            new_data = Data(new_bytes, len(new_bytes))
            return new_data

        def uncompressed_byte_size(self, data, values):
            # check number of samples
            if values[self.idx_nblocks] > self.nblocks.maximum:
                raise RPLError(
                    "The number of samples is superior to authorized values" +
                    " for packet {0}: {1} with maximum {2}".format(
                        self.name,
                        values[self.idx_nblocks],
                        self.nblocks.maximum,
                    )
                )

            # check number of channels
            if values[self.idx_nchannels] > self.nchannels.maximum:
                raise RPLError(
                    "The number of samples is superior to authorized values" +
                    " for packet {0}: {1} with maximum {2}".format(
                        self.name,
                        values[self.idx_nchannels],
                        self.nchannels.maximum,
                    )
                )
            return (
                values[self.idx_nblocks] * values[self.idx_nchannels] *
                self.block_size
            )

    return TDS


def create_packet(Packet):
    """
    Given the base class of the packets to use, add special methods into the
    new class to be able to treat the compressed packet as an existing one.
    """
    # get the base class for TDS packets
    TDS = get_base_tds(Packet)

    # TM_TDS_SCIENCE_NORMAL_RSWF
    class NormalRSWF(TDS):
        DATA_START = "PA_TDS_COMP_DATA_BLK"
        DATA_COUNT = "PA_TDS_N_RSWF_C_BLK_NR"
        START = "PA_TDS_COMPUTED_CRC"
        NBLOCK = "PA_TDS_SAMPS_PER_CH"
        NCHANNEL = "PA_TDS_NUM_CHANNELS"
        NC_DATA_START = "PA_TDS_RSWF_DATA_BLK"

    NormalRSWF("TM_TDS_SCIENCE_NORMAL_RSWF_C")

    # TM_TDS_SCIENCE_SBM1_RSWF
    class SBM1RSWF(NormalRSWF):
        DATA_COUNT = "PA_TDS_S1_RSWF_C_BLK_NR"

    SBM1RSWF("TM_TDS_SCIENCE_SBM1_RSWF_C")

    # TM_TDS_SCIENCE_NORMAL_TSWF
    class NormalTSWF(TDS):
        DATA_START = "PA_TDS_COMP_DATA_BLK"
        DATA_COUNT = "PA_TDS_N_TSWF_C_BLK_NR"
        START = "PA_TDS_COMPUTED_CRC"
        NBLOCK = "PA_TDS_SAMPS_PER_CH"
        NCHANNEL = "PA_TDS_NUM_CHANNELS"
        NC_DATA_START = "PA_TDS_TSWF_DATA_BLK"

    NormalTSWF("TM_TDS_SCIENCE_NORMAL_TSWF_C")

    # TM_TDS_SCIENCE_SBM2_TSWF
    class SBM2TSWF(NormalTSWF):
        DATA_COUNT = "PA_TDS_S2_TSWF_C_BLK_NR"

    SBM2TSWF("TM_TDS_SCIENCE_SBM2_TSWF_C")

    # TM_TDS_SCIENCE_NORMAL_MAMP
    class NormalMAMP(TDS):
        DATA_START = "PA_TDS_COMP_DATA_BLK"
        DATA_COUNT = "PA_TDS_MAMP_C_DATA_BLK_NR"
        START = "PA_TDS_COMPUTED_CRC"
        NBLOCK = "PA_TDS_MAMP_SAMP_PER_CH"
        NCHANNEL = "PA_TDS_MAMP_NUM_CH"
        NC_DATA_START = "PA_TDS_MAMP_DATA_BLK"

    NormalMAMP("TM_TDS_SCIENCE_NORMAL_MAMP_C")

    # TM_TDS_SCIENCE_LFM_CWF
    class LFMCWF(TDS):
        DATA_START = "PA_TDS_COMP_DATA_BLK"
        DATA_COUNT = "PA_TDS_LFM_CWF_C_BLK_NR"
        START = "PA_TDS_COMPUTED_CRC"
        NBLOCK = "PA_TDS_LFR_CWF_SAMPS_PER_CH"
        NCHANNEL = "PA_TDS_LFM_CWF_CH_NR"
        NC_DATA_START = "PA_TDS_LFM_CWF_DATA_BLK"

    LFMCWF("TM_TDS_SCIENCE_LFM_CWF_C")

    # TM_TDS_SCIENCE_LFM_RSWF
    class LFMRSWF(TDS):
        DATA_START = "PA_TDS_COMP_DATA_BLK"
        DATA_COUNT = "PA_TDS_LFM_RSWF_C_BLK_NR"
        START = "PA_TDS_COMPUTED_CRC"
        NBLOCK = "PA_TDS_LFR_SAMPS_PER_CH"
        NCHANNEL = "PA_TDS_LFM_RSWF_CH_NR"
        NC_DATA_START = "PA_TDS_LFM_RSWF_DATA_BLK"

    LFMRSWF("TM_TDS_SCIENCE_LFM_RSWF_C")


# connect to the signal of the PacketParser to generate handlers of compressed packets
PacketParser.reseted.connect(create_packet)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

try:
    from poppy.core.generic.metaclasses import Singleton
except:
    print('POPPy framework is not installed!')

__all__ = ['PacketCache']


class PacketCache(metaclass=Singleton):

    def __init__(self):
        # Cache for IDB
        self.idb = {}

        # Cache for packet max_sid values
        self.max_sid = {}

        # Cache for packet_id values
        self.packet_ids = {}

        # Cache for PALISADE metadata query result
        self.palisade_metadata_query = {}

        # Cache for PALISADE metadata dictionary
        self.palisade_metadata_dict = {}

        # Cache for latest PALISADE IDB version
        self.latest_palisade_version = None

        # Cache for transfer function
        self.transfer_function = {}




cdef class Parameter:
    def __cinit__(
        Parameter self,
        str name,
        uint32_t byte,
        uint32_t bit,
        uint32_t group,
        uint32_t block_size,
        uint32_t counter,
        uint32_t length,
        str type,
        double maximum,
    ):
        self.name = name
        self.byte = byte
        self.bit = bit
        self.group = group
        self.block_size = block_size
        self.length = length
        self.type = type
        self.maximum = maximum
        self.counter = counter

    def __repr__(Parameter self):
        return (
            "Parameter(" +
            "name={0},".format(self.name) +
            "byte={0},".format(self.byte) +
            "bit={0},".format(self.bit) +
            "length={0},".format(self.length) +
            "type={0},".format(self.type) +
            ")"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Testing roc.rpl.time.py."""


from roc.rpl.time import Time
import numpy as np
from datetime import datetime

TIME_INSTANCE = Time()

def test_cuc_to_nanosec():
    """Test roc.rpl.time.Time.cuc_to_nanosec function."""
    # Create input RPW CUC time
    in_time = np.empty([3, 2], dtype=np.uint32)
    # RPW CUC time at 2000-01-02 00:00:00.0762939453125
    in_time[0, :] = [86400, 5000]
    # RPW CUC time at 2010-01-02 00:00:00.9918212890625
    in_time[1, :] = [315619200, 65000]
    # RPW CUC time at 2017-01-02 00:00:00.0030517578125
    in_time[2, :] = [536630400, 200]

    # Create expected time
    exp_time = np.empty([3], dtype=np.uint64)
    exp_time[0] = 86400076293945
    exp_time[1] = 315619200991821289
    exp_time[2] = 536630400003051757

    out_time = Time.cuc_to_nanosec(in_time)

    for i, etime in enumerate(exp_time):
        # Assert at the microsecond resolution
        assert int(out_time[i]/1000.) == int(etime/1000.)

def test_cuc_to_microsec():
    """Test roc.rpl.time.Time.cuc_to_microsec function."""
    # Create input RPW CUC time
    in_time = np.empty([3, 2], dtype=np.uint32)
    # RPW CUC time at 2000-01-02 00:00:00.0762939453125
    in_time[0, :] = [86400, 5000]
    # RPW CUC time at 2010-01-02 00:00:00.9918212890625
    in_time[1, :] = [315619200, 65000]
    # RPW CUC time at 2017-01-02 00:00:00.0030517578125
    in_time[2, :] = [536630400, 200]

    # Create expected time
    exp_time = np.empty([3], dtype=np.uint64)
    exp_time[0] = 86400076294
    exp_time[1] = 315619200991821
    exp_time[2] = 536630400003052

    out_time = Time.cuc_to_microsec(in_time)

    for i, etime in enumerate(exp_time):
        # Assert at the microsecond resolution
        assert int(out_time[i]/1000.) == int(etime/1000.)

def test_obt_to_tt2000_nospice():
    """
    Test roc.rpl.time.Time.obt_to_utc function
    with to_tt2000=True and no_spice=True options.
    """
    # Create input RPW CUC time
    in_time = np.empty([3, 2], dtype=np.uint32)
    # RPW CUC time at 2000-01-02 00:00:00.0762939453125
    in_time[0, :] = [86400, 5000]
    # RPW CUC time at 2010-01-02 00:00:00.9918212890625
    in_time[1, :] = [315619200, 65000]
    # RPW CUC time at 2017-01-02 00:00:00.0030517578125
    in_time[2, :] = [536630400, 200]

    # Create expected time
    exp_time = np.empty([3], dtype=int)
    exp_time[0] = 43264260293
    exp_time[1] = 315576065175821
    exp_time[2] = 536587264187051

    TIME_INSTANCE.no_spice = True
    out_time = TIME_INSTANCE.obt_to_utc(in_time, to_tt2000=True)

    for i, etime in enumerate(exp_time):
        # Assert at the microsecond resolution
        assert int(out_time[i]/1000.0) == etime

def test_cuc_to_scet():
    """
    Test roc.rpl.time.Time.cuc_to_scet.

    :return:
    """

    # Create input RPW CUC time
    in_time = np.empty([2, 2], dtype=np.uint32)
    in_time[0, :] = [549570899, 64512]
    in_time[1, :] = [315619200, 65000]

    # Create expected time
    exp_time = np.empty([2], dtype=np.float64)
    exp_time[0] = 549570899.984375
    exp_time[1] = 315619200.9918213

    out_time = Time.cuc_to_scet(in_time)

    for i, etime in enumerate(exp_time):
        assert out_time[i] == etime


def test_obt_to_utc_nospice():
    """
    Test roc.rpl.time.Time.obt_to_utc without using SPICE.

    :return:
    """

    expected = datetime(2000, 1, 1, 0, 0, 0, 488)

    input = (0, 32, 1)

    TIME_INSTANCE.no_spice = True
    utc_time = TIME_INSTANCE.obt_to_utc([input], to_datetime=True)[0]

    assert utc_time == expected



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.generic.dot_dict import DotDict
from poppy.core.logger import logger

__all__ = ['get_raw_data']

def get_raw_data(pipeline):
    try:
        raw_data = {
             'packet_list': [{'binary': binary}
                             for binary in pipeline.args.binary],
             'idb_version': pipeline.args.idb_version[0],
             'idb_source': pipeline.args.idb_source[0],
         }
        return raw_data
    except:
        # If not defined as input argument, then assume that it is already
        # defined as target input
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest

from roc.rpl.packet_structure.data import Data


class TestRPLExtraction(unittest.TestCase):
    """
    To test the parameter extraction from bytearray.
    """

    def setUp(self):
        """
        Create the bytearray from where to read the data and store it.
        """
        # keep a reference, since it is C pointer inside Data
        self.byte_data = bytes(bytearray.fromhex("{:x}".format(4194967196)))

        # create data object
        self.data = Data(self.byte_data, len(self.byte_data))

    def test_unsigned_char(self):
        """
        Test the conversion for unsigned char at different positions.
        """
        self.assertEqual(
            self.data.u8p(0, 0, 8),
            250,
        )
        self.assertEqual(
            self.data.u8p(1, 0, 8),
            10,
        )
        self.assertEqual(
            self.data.u8p(2, 0, 8),
            30,
        )
        self.assertEqual(
            self.data.u8p(3, 0, 8),
            156,
        )
        self.assertEqual(
            self.data.u8p(2, 5, 6),
            52,
        )

    def test_unsigned_short(self):
        """
        Test the conversion for unsigned short at different positions.
        """
        self.assertEqual(
            self.data.u16p(0, 0, 16),
            64010,
        )
        self.assertEqual(
            self.data.u16p(2, 0, 16),
            7836,
        )
        self.assertEqual(
            self.data.u16p(1, 5, 6),
            16,
        )

    def test_unsigned_long(self):
        """
        Test the conversion for unsigned long at different positions.
        """
        self.assertEqual(
            self.data.u32p(0, 0, 32),
            4194967196,
        )

    def test_signed_short(self):
        self.assertEqual(
            self.data.i16p(0, 0, 16),
            -1526,
        )

    def test_signed_long(self):
        """
        Test read as signed long.
        """
        self.assertEqual(
            self.data.i32p(0, 0, 32),
            -100000100,
        )

    def test_signed_char(self):
        """
        Test the conversion for signed char at different positions.
        """
        self.assertEqual(
            self.data.i8p(0, 0, 8),
            -6,
        )
        self.assertEqual(
            self.data.i8p(1, 0, 8),
            10,
        )
        self.assertEqual(
            self.data.i8p(2, 0, 8),
            30,
        )
        self.assertEqual(
            self.data.i8p(3, 0, 8),
            -100,
        )
        self.assertEqual(
            self.data.i8p(2, 5, 6),
            52,
        )

    def test_float_float(self):
        """
        Test the value read for floats.
        """
        self.assertAlmostEqual(
            self.data.f32p(0, 0, 32),
            -1.79289449589817e+35,
        )



from libc.stdint cimport uint8_t
from libc.stdint cimport uint16_t

cdef class Header:
    cdef public uint8_t ccsds_version_number
    cdef public uint8_t packet_type
    cdef public uint8_t data_field_header_flag
    cdef public uint8_t process_id
    cdef public uint8_t packet_category
    cdef public uint8_t segmentation_grouping_flag
    cdef public uint16_t sequence_cnt
    cdef public uint16_t packet_length

    cpdef tuple to_tuple(Header self)
    cpdef dict to_dict(Header self)


cdef class TCHeader:
    cdef public uint8_t ccsds_secondary_header_flag
    cdef public uint8_t pus_version
    cdef public uint8_t ack_execution_completion
    cdef public uint8_t ack_execution_progress
    cdef public uint8_t ack_execution_start
    cdef public uint8_t ack_acceptance
    cdef public uint8_t service_type
    cdef public uint8_t service_subtype
    cdef public uint8_t source_id

    cpdef tuple to_tuple(TCHeader self)
    cpdef dict to_dict(TCHeader self)


cdef class TMHeader:
    cdef public uint8_t spare_1
    cdef public uint8_t pus_version
    cdef public uint8_t spare_2
    cdef public uint8_t service_type
    cdef public uint8_t service_subtype
    cdef public uint8_t destination_id
    cdef public tuple time

    cpdef tuple to_tuple(TMHeader self)
    cpdef dict to_dict(TMHeader self)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import glob

from distutils.command.build_ext import build_ext
from distutils.core import Extension
from distutils.errors import CCompilerError
from distutils.errors import DistutilsExecError
from distutils.errors import DistutilsPlatformError

# import numpy and cython
from Cython.Build import cythonize
import numpy

# the path for rice module and rpl
RICE_PATH = os.path.join('roc', 'rpl', 'rice')

# path to packet_structure for PacketParser
RPL_PACKET_STRUCTURE_PATH = os.path.join('roc', 'rpl', 'packet_structure')

# Path to packet_parser
RPL_PACKET_PARSER_PATH = os.path.join('roc', 'rpl', 'packet_parser', 'parser')

# C Extensions
with_extensions = os.getenv('RPL_EXTENSIONS', None)

if with_extensions == '1' or with_extensions is None:
    with_extensions = True

if with_extensions == '0' or hasattr(sys, 'pypy_version_info'):
    with_extensions = False

extensions = []
if with_extensions:
    extensions = list()
    extensions += cythonize(
        Extension(
            'roc.rpl.packet_structure.header',
            ['roc/rpl/packet_structure/header.pyx', 'roc/rpl/packet_structure/type_conversion.c'],
            include_dirs=[numpy.get_include(), RPL_PACKET_STRUCTURE_PATH],
        ),
        language_level='3',
    )
    extensions += cythonize(
        Extension(
            'roc.rpl.packet_structure.parameter',
            ['roc/rpl/packet_structure/parameter.pyx', 'roc/rpl/packet_structure/type_conversion.c'],
            include_dirs=[numpy.get_include(), RPL_PACKET_STRUCTURE_PATH],
        ),
        language_level='3',
    )
    extensions += cythonize(
        Extension(
            'roc.rpl.packet_structure.data',
            ['roc/rpl/packet_structure/data.pyx', 'roc/rpl/packet_structure/type_conversion.c'],
            include_dirs=[numpy.get_include(), RPL_PACKET_STRUCTURE_PATH],
        ),
        language_level='3',
    )
    extensions += cythonize(
        Extension(
            'roc.rpl.packet_parser.parser.identify_packets',
            sources=['roc/rpl/packet_parser/parser/identify_packets.pyx',
                     'roc/rpl/packet_structure/type_conversion.c'],
            include_dirs=[numpy.get_include(), RPL_PACKET_PARSER_PATH,
                          RPL_PACKET_STRUCTURE_PATH]
        ),
        language_level='3',
    )
    extensions += cythonize(
        Extension(
            'roc.rpl.packet_parser.parser.parse_packets',
            sources=['roc/rpl/packet_parser/parser/parse_packets.pyx',
                     'roc/rpl/packet_structure/type_conversion.c'],
            include_dirs=[numpy.get_include(), RPL_PACKET_PARSER_PATH,
                          RPL_PACKET_STRUCTURE_PATH]
        ),
        language_level='3',
    )

    # now add an extension for compiling the rice compression library
    extensions += cythonize(
        Extension(
            'roc.rpl.rice.rice',  # this will be the name of the module
            sources=[os.path.join(RICE_PATH, 'rice.pyx')] +
                    [fn for fn in glob.glob(os.path.join(RICE_PATH, '*.cpp'))
                     if not os.path.basename(fn) == 'rice.cpp'],
            include_dirs=[numpy.get_include(), RICE_PATH, RPL_PACKET_STRUCTURE_PATH],
            language='c++',
            define_macros=[('GAUSS', '1')],
        ),
        language_level='3',
    )


class BuildFailed(Exception):

    pass


class ExtBuilder(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except (DistutilsPlatformError, FileNotFoundError):
            print('************************************************************')
            print('Cannot compile RPL Cython module!')
            print('************************************************************')

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError, ValueError):
            print('************************************************************')
            print('Cannot compile Cython module!')
            print('************************************************************')


def build(setup_kwargs):
    """
    This function is mandatory in order to build the extensions.
    """
    setup_kwargs.update(
        {'ext_modules': extensions, 'cmdclass': {'build_ext': ExtBuilder}}
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Module for the special handling of compressed data inside LFR packets.
"""

from roc.rpl.exceptions import RPLError
from roc.rpl.packet_parser.packet_parser import PacketParser
from .utils import get_base_packet

__all__ = []

def get_base_lfr(Packet):
    """
    Generate the base class to use for compressed packets in the case of LFR.
    """
    # base class for compressed packet
    Base = get_base_packet(Packet)

    # base class for the compressed packets of LFR
    class LFR(Base):
        """
        Base class for the compressed packets of LFR. Define the special
        behaviour of the treatment for those packets, that should mimic the non
        compressed (NC) ones.
        """
        def uncompressed_byte_size(self, data, values):
            """
            To get the size in bytes of the uncompressed data.
            """
            # get the number of blocks and return the size in bytes of the
            # uncompressed data
            if values[self.idx_nblocks] > self.nblocks.maximum:
                raise RPLError(
                    'The number of blocks is superior to authorized values' +
                    ' for packet {0}: {1} with maximum {2}'.format(
                        self.name,
                        values[self.idx_nblocks],
                        self.nblocks.maximum,
                    )
                )
            return values[self.idx_nblocks] * self.block_size

    return LFR


def create_packet(Packet):
    """
    Given the base class of the packets to use, add special methods into the
    new class to be able to treat the compressed packet as an existing one.
    """
    # get the base class for LFR
    LFR = get_base_lfr(Packet)

    # base class for the compressed packets of LFR
    class BurstCWFF2(LFR):
        DATA_START = 'PA_LFR_COMP_DATA_BLK'
        DATA_COUNT = 'PA_LFR_B_CWF_F2_C_BLK_NR'
        START = 'PA_LFR_COMPUTED_CRC'
        NBLOCK = 'PA_LFR_CWF_BLK_NR'
        NC_DATA_START = 'PA_LFR_SC_V_F2'

    BurstCWFF2('TM_LFR_SCIENCE_BURST_CWF_F2_C')

    class NormalCWFF3(LFR):
        DATA_START = 'PA_LFR_COMP_DATA_BLK'
        DATA_COUNT = 'PA_LFR_N_CWF_F3_C_BLK_NR'
        START = 'PA_LFR_COMPUTED_CRC'
        NBLOCK = 'PA_LFR_CWF3_BLK_NR'
        NC_DATA_START = 'PA_LFR_SC_V_F3'

    NormalCWFF3('TM_LFR_SCIENCE_NORMAL_CWF_F3_C')
    NormalCWFF3('TM_DPU_SCIENCE_BIAS_CALIB_C')

    class NormalCWFLongF3(NormalCWFF3):
        NBLOCK = 'PA_LFR_CWFL3_BLK_NR'

    NormalCWFLongF3('TM_LFR_SCIENCE_NORMAL_CWF_LONG_F3_C')
    NormalCWFLongF3('TM_DPU_SCIENCE_BIAS_CALIB_LONG_C')

    class NormalSWFF0(LFR):
        DATA_START = 'PA_LFR_COMP_DATA_BLK'
        DATA_COUNT = 'PA_LFR_N_SWF_F0_C_BLK_NR'
        START = 'PA_LFR_COMPUTED_CRC'
        NBLOCK = 'PA_LFR_SWF_BLK_NR'
        NC_DATA_START = 'PA_LFR_SC_V_F0'

    NormalSWFF0('TM_LFR_SCIENCE_NORMAL_SWF_F0_C')

    class NormalSWFF1(LFR):
        DATA_START = 'PA_LFR_COMP_DATA_BLK'
        DATA_COUNT = 'PA_LFR_N_SWF_F1_C_BLK_NR'
        START = 'PA_LFR_COMPUTED_CRC'
        NBLOCK = 'PA_LFR_SWF_BLK_NR'
        NC_DATA_START = 'PA_LFR_SC_V_F1'

    NormalSWFF1('TM_LFR_SCIENCE_NORMAL_SWF_F1_C')

    class NormalSWFF2(LFR):
        DATA_START = 'PA_LFR_COMP_DATA_BLK'
        DATA_COUNT = 'PA_LFR_N_SWF_F2_C_BLK_NR'
        START = 'PA_LFR_COMPUTED_CRC'
        NBLOCK = 'PA_LFR_SWF_BLK_NR'
        NC_DATA_START = 'PA_LFR_SC_V_F2'

    NormalSWFF2('TM_LFR_SCIENCE_NORMAL_SWF_F2_C')

    class SBM1CWFF1(LFR):
        DATA_START = 'PA_LFR_COMP_DATA_BLK'
        DATA_COUNT = 'PA_LFR_S1_CWF_F1_C_BLK_NR'
        START = 'PA_LFR_COMPUTED_CRC'
        NBLOCK = 'PA_LFR_CWF_BLK_NR'
        NC_DATA_START = 'PA_LFR_SC_V_F1'

    SBM1CWFF1('TM_LFR_SCIENCE_SBM1_CWF_F1_C')

    class SBM2CWFF2(LFR):
        DATA_START = 'PA_LFR_COMP_DATA_BLK'
        DATA_COUNT = 'PA_LFR_S2_CWF_F2_C_BLK_NR'
        START = 'PA_LFR_COMPUTED_CRC'
        NBLOCK = 'PA_LFR_CWF_BLK_NR'
        NC_DATA_START = 'PA_LFR_SC_V_F2'

    SBM2CWFF2('TM_LFR_SCIENCE_SBM2_CWF_F2_C')


# connect to the signal of the PacketParser to generate handlers of compressed packets
PacketParser.reseted.connect(create_packet)



import numpy as np
cimport numpy as np

from libc.stdint cimport int8_t
from libc.stdint cimport uint8_t
from libc.stdint cimport int16_t
from libc.stdint cimport uint16_t
from libc.stdint cimport int32_t
from libc.stdint cimport uint32_t
from cpython cimport bool
from .data cimport to_float
from .data cimport to_signed_char
from .data cimport to_signed_short
from .data cimport to_signed_long
from .parameter cimport Parameter
from .header cimport TCHeader
from .header cimport TMHeader
from .header cimport Header

cpdef tuple header_foot_offset(uint8_t packet_type):
    """
    Get offset according to packet type.
    All packets have an offset of the application and source data at a
    different position according to the type of the packet, so the header
    offset take of this.
    Moreover, TC commands have a spare not defined as a parameter at the end
    of the packet called CRC, of 2 bytes size, so we had it for TC packets
    """
    cdef:
        Py_ssize_t header_offset
        Py_ssize_t foot_offset

    # If TC packet
    if packet_type == 1:
        header_offset = 10
        foot_offset = 2
    # If TM packet
    else:
        header_offset = 16
        foot_offset = 0

    return header_offset, foot_offset

cdef class Data:
    """
    Wrapper class around bytes array to extract simple values, headers,
    parameters and block parameters.
    """
    def __cinit__(Data self, bytes data, uint32_t size):
        self.ref = data
        self.data = <uint8_t*> self.ref
        self.size = size

    def __str__(self):
        return (
                "Data(size=" + str(self.size) +
                ", buf=" + str(self.data[:self.size]) + ")"
        )

    property data_size:
        def __get__(self):
            return self.size

        def __set__(self, size):
            self.size = size

    property data_buffer:
        def __get__(self):
            return <bytes> self.data[:self.size]

    cpdef to_bytes(Data self):
        """
        Return the data into a bytearray that can be manipulated directly in
        Python without using C code.
        """
        return bytes(self.data[:self.size])

    cpdef extract_parameter(Data self, Parameter parameter, uint32_t offset):
        """
        Given a parameter, extract its value at the given offset.
        """
        # get the value of the parameter
        if parameter.type == "uint8":
            return self.u8(
                parameter.byte + offset,
                parameter.bit,
                parameter.length,
            )
        elif parameter.type == "uint16":
            return self.u16(
                parameter.byte + offset,
                parameter.bit,
                parameter.length,
            )
        elif parameter.type == "uint32":
            return self.u32(
                parameter.byte + offset,
                parameter.bit,
                parameter.length,
            )
        elif parameter.type == "int8":
            return self.i8(
                parameter.byte + offset,
                parameter.bit,
                parameter.length,
            )
        elif parameter.type == "int16":
            return self.i16(
                parameter.byte + offset,
                parameter.bit,
                parameter.length,
            )
        elif parameter.type == "int32":
            return self.i32(
                parameter.byte + offset,
                parameter.bit,
                parameter.length,
            )
        elif parameter.type == "float32":
            return self.f32(
                parameter.byte + offset,
                parameter.bit,
                parameter.length,
            )
        elif parameter.type == "time":
            return self.time(
                parameter.byte + offset,
                parameter.bit,
                parameter.length,
            )

    cpdef Header extract_header(Data self):
        """
        Extract header information from the packet data.
        """
        header = Header(
            ccsds_version_number=self.u8(0, 0, 3),
            packet_type=self.u8(0, 3, 1),
            data_field_header_flag=self.u8(0, 4, 1),
            process_id=self.u8(0, 5, 7),
            packet_category=self.u8(1, 4, 4),
            segmentation_grouping_flag=self.u8(2, 0, 2),
            sequence_cnt=self.u16(2, 2, 14),
            packet_length=self.u16(4, 0, 16),
        )
        return header

    cpdef TCHeader extract_tc_header(Data self):
        """
        Extract TC data field header.
        """
        header = TCHeader(
            ccsds_secondary_header_flag=self.u8(
                6, 0, 1
            ),
            pus_version=self.u8(6, 1, 3),
            ack_execution_completion=self.u8(6, 4, 1),
            ack_execution_progress=self.u8(6, 5, 1),
            ack_execution_start=self.u8(6, 6, 1),
            ack_acceptance=self.u8(6, 7, 1),
            service_type=self.u8(7, 0, 8),
            service_subtype=self.u8(8, 0, 8),
            source_id=self.u8(9, 0, 8),
        )
        return header

    cpdef TMHeader extract_tm_header(Data self):
        """
        Extract TM data field header.
        """
        header = TMHeader(
            spare_1=self.u8(6, 0, 1),
            pus_version=self.u8(6, 1, 3),
            spare_2=self.u8(6, 4, 4),
            service_type=self.u8(7, 0, 8),
            service_subtype=self.u8(8, 0, 8),
            destination_id=self.u8(9, 0, 8),
            time=self.time(10, 0, 48),
        )
        return header

    cpdef float f32p(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ):
        return self.f32(start, offset, length)

    cdef float f32(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ) nogil:
        cdef:
            float result
            uint32_t tmp

        # get the value with uint32_t
        tmp = self.u32(start, offset, length)

        result = to_float(tmp)
        return result

    cpdef uint32_t u32p(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ):
        return self.u32(
            start,
            offset,
            length,
        )

    cdef uint32_t u32(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ) nogil:
        """
        Extract an uint32_t from a byte array, with the starting byte in
        the array, the offset in bit from the left most position in the array
        and a bit length of length.
        """
        cdef:
            uint32_t byte1
            uint32_t byte2
            uint32_t byte3
            uint32_t result

        # compute the first two bytes from the offset
        byte1 = <uint32_t> self.u8(start + 3, offset, 8)
        byte2 = <uint32_t> self.u8(start + 2, offset, 8)
        byte3 = <uint32_t> self.u8(start + 1, offset, 8)
        result = <uint32_t> self.u8(start, offset, 8)

        # compute the short integer
        result = result << 24
        byte3 = byte3 << 16
        byte2 = byte2 << 8
        result = byte1 | byte2 | byte3 | result

        # short it by the length
        result = result >> (32 - length)
        return result

    cpdef int32_t i32p(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ):
        cdef:
            int32_t result
        result = self.i32(start, offset, length)
        return result

    cdef int32_t i32(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ) nogil:
        cdef:
            uint32_t tmp
            int32_t result
        # get as unsigned
        tmp = self.u32(start, offset, length)

        # reinterpret as signed
        result = to_signed_long(tmp)
        return result

    cpdef uint16_t u16p(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ):
        return self.u16(
            start,
            offset,
            length,
        )

    cdef uint16_t u16(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ) nogil:
        cdef:
            uint16_t byte2
            uint16_t result

        # compute the first two bytes from the offset
        byte2 = <uint16_t> self.u8(start + 1, offset, 8)
        result = <uint16_t> self.u8(start, offset, 8)

        # compute the short integer
        result = result << 8
        result = byte2 | result

        # short it by the length
        result = result >> (16 - length)
        return result

    cpdef int16_t i16p(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ):
        return self.i16(start, offset, length)

    cdef int16_t i16(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ) nogil:
        cdef:
            uint16_t tmp
            int16_t result
        tmp = self.u16(start, offset, length)
        result = to_signed_short(tmp)
        return result

    cpdef uint8_t u8p(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length
    ):
        return self.u8(start, offset, length)

    cdef uint8_t u8(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length
    ) nogil:
        """
        We store the intermediate result of shifting the bits because cython is
        doing an implicit conversion to an 'higher' size integer if the shift
        is bigger than a byte.
        """
        cdef:
            uint8_t result
            uint8_t tmp

        # start by reading the entire byte
        with nogil:
            result = self.data[start] << offset
            tmp = self.data[start + 1] >> (8 - offset)
            result = result | tmp

            # and hide the last part
            result = result >> (8 - length)
        return result

    cpdef int8_t i8p(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length
    ):
        return self.i8(start, offset, length)

    cdef int8_t i8(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length
    ) nogil:
        """
        We store the intermediate result of shifting the bits because cython is
        doing an implicit conversion to an 'higher' size integer if the shift
        is bigger than a byte.

        Do not do a lot of smart things with the sign bit since if we want to
        read a signed value, we need to read 8 bit, so the length should be 8.
        So I think that it should be always used with the length equals to 8.
        """
        cdef:
            uint8_t tmp
            int8_t result

        tmp = self.u8(start, offset, length)
        result = to_signed_char(tmp)
        return result

    cdef tuple time(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length,
    ):
        cdef:
            uint32_t coarse
            uint16_t fine
            uint8_t flag

        # get the flag
        flag = self.u8(start, 0, 1)

        # read the coarse part
        coarse = self.u32(start, 1, 31)

        # read the fine part
        fine = self.u16(start + 4, 0, 16)

        return (coarse, fine, flag)


    cpdef tuple timep(
            Data self,
            uint32_t start,
            uint32_t offset,
            uint32_t length
    ):
        return self.time(start, offset, length)


    cpdef tuple extract_block(
            Data self,
            Py_ssize_t offset,
            list parameters,
            uint32_t counter
    ):
        cdef:
            Py_ssize_t j, k
            Py_ssize_t limit
            uint32_t group_size
            uint32_t block_size
            list buff
            list values

            Parameter block_parameter

        # init the list of parameters and their data
        values = []

        # size of the group
        group_size = parameters[0].group

        # empty list, a list of values for each parameter in the block, since
        # multiple values are present for a parameter
        buff = [[] for x in range(group_size)]

        # compute the block offset (the block size in bit, including the any spare
        # present in the block, but not present in the parameter list)
        block_size = parameters[0].block_size

        # loop over parameters for the block size
        for j in range(counter):
            # loop over parameters in group
            for k in range(group_size):
                # get the parameter
                block_parameter = parameters[k]

                # get the value of the parameter
                value = self.extract_parameter(block_parameter, offset)

                # add the value to the buffer
                buff[k].append(value)

            # add the block to the offset
            offset += block_size

        # since the byte position of parameters from the database already
        # take into account one block size for their definition, we need
        # to delete the last offset of the parameters to be correctly
        # aligned with the size provided by the packet himself
        offset -= block_size

        # add values to the result
        for k in range(group_size):
            # get current parameter
            block_parameter = parameters[k]

            # create data array with the good type
            results = np.empty(
                counter,
                dtype=block_parameter.type,
            )

            # load the buffer data in the result array
            results[:] = buff[k][:]

            # add data to values
            values.append(results)

        return values, offset, group_size

    cpdef tuple extract_parameters(
            Data self,
            Py_ssize_t offset,
            list parameters,
    ):
        """
        Given the parameters to extract, get them from byte data, and take into
        account the blocks.
        """
        cdef:
            Py_ssize_t i

            uint32_t nmax
            uint32_t group_size
            uint32_t counter_value
            uint32_t data_field_offset

            bool in_block

            list values
            list buff

            Parameter block_parameter
            Parameter last_param

        # init some variables
        values = []

        # loop on parameters
        i = 0
        nmax = len(parameters)
        while (i < nmax):
            # check we are in block
            in_block = parameters[i].group > 0

            # if not in a block, read the parameters with the given offset (header
            # and eventually blocks of data)
            if not in_block:
                # store the last parameter if not in block and the value
                counter_parameter = parameters[i]

                # get the value of the parameter
                value = self.extract_parameter(parameters[i], offset)

                # add the value to the result
                values.append(value)

                # next parameter
                i += 1

            else:
                # case where the counter is from a parameter, and is by definition
                # the last extracted value
                if parameters[i].counter == 0:
                    # the last stored value is the counter
                    counter_value = value

                    # compute the maximal value of the array, depending on the
                    # maximal value of the counter
                    maximum = counter_parameter.maximum
                # case where the value of the counter is hardcoded in the IDB, the
                # IDB database of the store this information in the counter field
                # of the parameter, and the counter and maximum are the same
                else:
                    counter_value = parameters[i].counter
                    maximum = parameters[i].counter

                # check the data size is compatible with expected limits from
                # the IDB definition
                if counter_value > maximum:
                    raise RPLDecommuteError(f"Bad value {counter_value} for counter parameter {counter_parameter.name} "
                                            f"with maximum {maximum}")

                # extract the block data from the information of parameters
                buff, offset, group_size = self.extract_block(
                    offset,
                    parameters[i:],
                    counter_value,
                )

                # add values for block into the values container
                values += buff

                # increment parameters
                i += group_size

        return tuple(values), offset


class RPLDecommuteError(Exception):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from sqlalchemy import func
from sqlalchemy.orm.exc import NoResultFound

try:
    from poppy.core.db.connector import Connector
    from poppy.core.logger import logger
except:
    print('POPPy framework is not installed!')

try:
    from roc.idb.models.palisade_metadata import PalisadeMetadata
except:
    print('roc.idb plugin is not installed!')

from roc.rpl.packet_parser.packet_cache import PacketCache

__all__ = ['latest_palisade_version',
           'query_palisade_metadata',
           'palisade_metadata']


@Connector.if_connected('MAIN-DB')
def latest_palisade_version():
    """
    Get latest palisade IDB version in the database

    :return: latest palisade IDB version
    """

    # Initialize output
    latest_palisade_version = None

    # get a database session
    session = Connector.manager['MAIN-DB'].session

    # Get/create PacketCache single instance
    packet_cache = PacketCache()

    # Check in packet_cache for latest version
    if packet_cache.latest_palisade_version:
        latest_palisade_version = packet_cache.latest_palisade_version
    else:
        # Get latest version from database
        try:
            logger.debug(f'Querying latest PALISADE IDB version from the database')
            result = session.query(func.max(PalisadeMetadata.palisade_version)).one()
        except:
            logger.exception('Get palisade latest version from the database has failed!')
        else:
            latest_palisade_version = result[0]
            # Store in packet cache
            packet_cache.latest_palisade_version = latest_palisade_version

    return latest_palisade_version


@Connector.if_connected('MAIN-DB')
def query_palisade_metadata(palisade_version=None):
    """
    Method to retrieve PALISADE IDB metadata from database.

    :param palisade_version: Version of the PALISADE IDB
    :return: packet PALISADE metadata as a result of the query
    """

    # Initialize output dict
    palisade_metadata = {}

    # get a database session
    session = Connector.manager['MAIN-DB'].session

    # Get/create PacketCache single instance
    packet_cache = PacketCache()

    # Get palisade version
    if not palisade_version:
        # Return palisade version
        palisade_version = latest_palisade_version()

    if (palisade_version
        and palisade_version in packet_cache.palisade_metadata_query):
        # Check if input parameters are already in the cache for the given palisade_version
        palisade_metadata = packet_cache.palisade_metadata_query[palisade_version]
    elif palisade_version:
        # Else get palisade_metadata from database
        try:
            logger.debug(f'Querying palisade metadata for version {palisade_version}')
            query = session.query(PalisadeMetadata)
            query = query.filter_by(palisade_version=palisade_version)
            results = query.all()
        except NoResultFound:
            logger.exception(f'No palisade metadata found in the database for version {palisade_version}')
        else:
            packet_cache.palisade_metadata_query[palisade_version] = results
            palisade_metadata = results

    return palisade_metadata

def palisade_metadata(palisade_version=None):
        """
        Method to retrieve PALISADE IDB metadata from database.

        :param palisade_version: Version of the PALISADE IDB
        :return: packet PALISADE metadata as a dictionary (keywords are the srdb and palisade id)
        """

        # output dictionary initialization
        palisade_metadata_dict = {}

        # Get palisade version
        if not palisade_version:
            # Return palisade version
            palisade_version = latest_palisade_version()

        # Get packet cache
        packet_cache = PacketCache()
        if (palisade_version and
            palisade_version in packet_cache.palisade_metadata_dict):
            # Check if already in the cache for the given palisade_version
            palisade_metadata_dict = packet_cache.palisade_metadata_dict[palisade_version]
        elif palisade_version:
            # If not in the cache, build palisade_metadata_dict dictionary
            palisade_metadata_query = query_palisade_metadata(palisade_version=palisade_version)

            if palisade_metadata_query:
                # Initialize output dictionary
                palisade_metadata_dict[palisade_version] = {}

                # And fill the output dictionary
                for row in palisade_metadata_query:
                    # Create two keyword entries in the dictionary:
                    # One for the palisade_id (row values are also returned into a dictionary)
                    palisade_metadata_dict[row.palisade_id] = {c.name: getattr(row, c.name) for c in row.__table__.columns}
                    # One for the srdb_id (row values are also returned into a dictionary)
                    palisade_metadata_dict[row.srdb_id] = {c.name: getattr(row, c.name) for c in row.__table__.columns}

                    packet_cache.palisade_metadata_dict[palisade_version] = palisade_metadata_dict

        return palisade_metadata_dict




try:
    from roc.idb.parsers.idb_parser import IDBParser
except:
    print('roc.idb framework is not installed!')

from roc.rpl import Time
from roc.rpl.packet_parser.extended_data import ExtendedData
from roc.rpl.constants import TM_PACKET_TYPE, \
    TC_PACKET_TYPE, INVALID_UTC_DATETIME, \
    UNKNOWN_UNIQUE_ID, UNKNOWN_ACK_STATE, INVALID_PACKET_HEADER, \
    OUTSIDE_RANGE_PACKET, VALID_PACKET, INVALID_PACKET_TIME

from roc.rpl.packet_parser import PacketParser

__all__ = ["identify_packets"]

cpdef tuple get_packet_id(object raw_binary_packet, object idb,
                          srdb_id=None, palisade_id=None):
    """
    Get packet ids (SRDB_ID and PALISADE_ID)

    :param raw_binary_packet: String containing the binary raw data of the packet (in hexa)
    :param idb: idb manager (from PacketParser class instance)
    :param srbd_id: packet SRDB ID (if passed, skip srdb id retrieval from database)
    :param palisade_id: packet PALISADE ID (if passed, skip srdb id retrieval from database)
    :return: srdb_id, palisade_id, packet_header, data_field_header
    """
    cdef:
        object byte_data

    # Convert hex to bytes
    byte_data = PacketParser.hex_to_bytes(raw_binary_packet)
    # Get IDs of the packet
    return ExtendedData(bytes(byte_data),
                        len(byte_data),
                        idb).get_packet_summary(srdb_id=srdb_id,
                                                palisade_id=palisade_id)


cdef dict identify_packet(int packet_idx,
                          dict packet_data,
                          object packet_parser,
                          start_time=None,
                          end_time=None):
    """
    Given a dict containing the binary data [and optionally the utc time] of the packet, return an identified packet
     and allocate space for data in the packet manager.

    :param packet_idx: Index of the current packet
    :param packet_data: dictionary containing packet data
    :param packet_parser: Instance of PacketParser class
    :param start_time: Only keep packet(s) created/executed after start_time (UTC)
    :param end_time: Only keep packet(s) created/executed before end_time (UTC)
    :return: dictionary with packet information (empty packet is returned if outside start_time/end_time)
    """
    cdef:
        unsigned long coarse
        unsigned long fine
        unsigned long expected_length
        unsigned long actual_length
        object comment

    # Make sure to have comment in packet_data dict
    comment = packet_data.get('comment')
    if comment is None:
        packet_data['comment'] = ''

    # Check if input packet has the required keyword
    packet_data['binary'] = packet_data.get('binary',
                                            PacketParser.get_binary(packet_data))
    if packet_data['binary'] is None:
        return packet_data

    # Store idb_source in packet_data
    packet_data['idb_source'] = packet_parser.idb_source

    # Store idb_version in packet_data
    packet_data['idb_version'] = packet_parser.idb_version

    # Retrieve packet SRDB_ID, PALISADE_ID using IDB version stored in the ROC database
    # (return the packet header and data_header at the same time)
    try:
        packet_data['srdb_id'], packet_data['palisade_id'], packet_data['header'], \
            packet_data['data_header'] = get_packet_id(packet_data["binary"], packet_parser.idb,
                                                       srdb_id=packet_data.get(
                                                           'srdb_id'),
                                                       palisade_id=packet_data.get('palisade_id'))
    except:
        msg = f'Packet #{packet_idx + 1} cannot be analysed! [{packet_data["binary"]}]'
        packet_data['comment'] = msg
        packet_data['status'] = INVALID_PACKET_HEADER
        return packet_data


    # skip the next steps if the name is unknown
    if packet_data['srdb_id'] is None:
        msg = f'Packet #{packet_idx + 1} cannot be identified! [{packet_data["binary"]}]. '
        packet_data['status'] = INVALID_PACKET_HEADER
        packet_data['comment'] += msg
        return packet_data

    # Compute APID
    try:
        packet_data['apid'] = IDBParser.compute_apid(packet_data['header'].process_id,
                                                     packet_data['header'].packet_category)
    except:
        msg = f'APID cannot be computed for {packet_data}. '
        packet_data['comment'] += msg

    # Get/create packet category
    packet_data['category'] = packet_data.get('category',
                                              packet_parser.palisade_metadata[packet_data['srdb_id']]['packet_category'])

    # Get UTC time of the packet generation if TM...
    if packet_data['header'].packet_type == TM_PACKET_TYPE:

        packet_data['type'] = 'TM'
        # Get packet scet or compute it
        packet_data['scet'] = packet_data.get('scet',
                                              Time().cuc_to_scet([packet_data['data_header'].time])[0])

        # Get utc_time or compute it for current TM packet
        try:
            packet_data['utc_time'] = Time().obt_to_utc(
                [packet_data['data_header'].time[:2]], to_datetime=True)[0]
        except:
            packet_data['utc_time'] = INVALID_UTC_DATETIME

        # Get time sync flag
        packet_data['sync_flag'] = (
            packet_data['data_header'].time[2] == 0)

       # Get/create on-board time using SPICE format
        coarse = packet_data['data_header'].time[0]
        fine = packet_data['data_header'].time[1]
        packet_data['obt_time'] = packet_data.get('obt_time',
                                                  f"1/{coarse}:{fine}")

    # If TC packet, then get execution time from tc report
    elif packet_data['header'].packet_type == TC_PACKET_TYPE:

        # Convert TC execution UTC time in datetime object
        packet_data['utc_time'] = packet_data.get('utc_time',
                                                  INVALID_UTC_DATETIME)
        packet_data['type'] = 'TC'
        packet_data['scet'] = None

        # Get TC acknowledgement status (acceptance and execution)
        packet_data['ack_exe_state'] = packet_data.get(
            'ack_exe_state', UNKNOWN_ACK_STATE)
        packet_data['ack_acc_state'] = packet_data.get(
            'ack_acc_state', UNKNOWN_ACK_STATE)

        # Get unique ID (if any)
        packet_data['unique_id'] = packet_data.get(
            'unique_id', UNKNOWN_UNIQUE_ID)

        # Get unique ID (if any)
        packet_data['sequence_name'] = packet_data.get('sequence_name', None)

    else:
        msg = f"Unknown packet_type {packet_data['header'].packet_type} " \
            f"for packet #{packet_idx} [{packet_data['srdb_id']}]. "
        packet_data['comment'] += msg
        packet_data['status'] = INVALID_PACKET_HEADER
        return packet_data

    # Get/create packet SHA
    packet_data['sha'] = PacketParser.get_packet_sha(packet_data)

    # Check that the packet length is as expected
    expected_length = packet_data['header'].packet_length + 7
    packet_data['length'] = expected_length
    actual_length = len(PacketParser.hex_to_bytes(packet_data['binary']))
    if expected_length != actual_length:
        msg = f"Packet {packet_data['palisade_id']} {{{packet_data['srdb_id']}}} on " \
            f"{packet_data['utc_time']} has not expected length: " \
            f"expected is {expected_length} and actual is {actual_length}. "
        packet_data['status'] = INVALID_PACKET_HEADER
        packet_data['comment'] += msg
        return packet_data

    # If input argument start_time has been provided...
    if (packet_data['utc_time'] != INVALID_UTC_DATETIME and
            start_time and start_time > packet_data['utc_time']):
        packet_data['status'] = OUTSIDE_RANGE_PACKET
        msg = (f"Packet {packet_data['palisade_id']} {{{packet_data['srdb_id']}}} on "
               f"{packet_data['utc_time']} is outside the time range. ")
        packet_data["comment"] += msg
        return packet_data

    # If input argument end_time has been provided...
    if (packet_data['utc_time'] != INVALID_UTC_DATETIME and
            end_time and end_time < packet_data['utc_time']):
        packet_data["status"] = OUTSIDE_RANGE_PACKET
        msg = (f"Packet {packet_data['palisade_id']} {{{packet_data['srdb_id']}}} on "
               f"{packet_data['utc_time']} is outside the time range. ")
        packet_data["comment"] += msg
        return packet_data

    if packet_data['utc_time'] == INVALID_UTC_DATETIME:
        packet_data["status"] = INVALID_PACKET_TIME
        packet_data['comment'] = f'Packet generation time cannot be converted into valid UTC time'
        return packet_data

    packet_data['status'] = VALID_PACKET

    return packet_data


cpdef list identify_packets(object packet_parser,
                            list packet_list,
                            start_time=None,
                            end_time=None):
    """
    Given a list of packets, analyze the binary content
    to identify the packets and allocate space for data.

    :param packet_parser: instance of PacketParser class
    :param packet_list: List of packets to identify
    :param start_time: Only keep packet(s) created/executed after start_time
    :param end_time: Only keep packet(s) created/executed before end_time
    :return: Store results in the PacketParser object
    """
    cdef:
        long packet_count
        Py_ssize_t packet_idx
        list analyzed_packet
        object palisade_version

    # init the class for managing packets for a given instance (do it for
    # each initialization to avoid problems)
    packet_parser.reset()

    # get number of packets to identify
    packet_count = len(packet_list)

    # Loop over packets
    analyzed_packet = [
        identify_packet(packet_idx, packet_list[packet_idx],
                        packet_parser,
                        start_time=start_time,
                        end_time=end_time)
        for packet_idx in range(packet_count)
    ]

    return analyzed_packet



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from roc.rpl.packet_parser.packet_parser import *
from roc.rpl.packet_parser.transfer_function import *
from roc.rpl.packet_parser.palisade import *
from roc.rpl.packet_parser.raw_data import *
from roc.rpl.packet_parser.transfer_function import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import uuid
import json

from roc.rpl.tasks.tools import get_raw_data

try:
    from poppy.core.logger import logger
    from poppy.core.task import Task
    from poppy.core.target import PyObjectTarget, FileTarget
    from poppy.core.tools.exceptions import MissingInput
    from poppy.core.db.connector import Connector
except:
    print('POPPy framework is not installed!')

__all__ = ['PacketsToJson']


class PacketsToJson(Task):
    """
    Task to print the input RPW packets.
    """

    plugin_name = 'roc.rpl'
    name = 'packets_to_json'

    def add_targets(self):
        self.add_input(identifier='raw_data',
                       value=get_raw_data(self.pipeline),
                       target_class=PyObjectTarget)

    def setup_inputs(self):
        # Get input target raw_data value
        try:
            self.raw_data = self.inputs['raw_data'].value
        except:
            raise MissingInput('raw_data input is missing')

        # If passed, retrieve path of the output JSON file
        self.json_file = self.pipeline.get('output_json_file',
                                               default=[None])[0]

        # Retrieve input boolean keywords
        self.print = self.pipeline.get('print',
                                            default=False)
        self.overwrite = self.pipeline.get('overwrite',
                                            default=False)

    def run(self):

        # Define task job ID (long and short)
        self.job_uuid = str(uuid.uuid4())
        self.job_id = f'PrintPacket-{self.job_uuid[:8]}'
        logger.debug(f'Task {self.job_id} is starting')
        try:
            self.setup_inputs()
        except:
            logger.exception(
                f'Initializing inputs has failed for {self.job_id}!')
            self.pipeline.exit()
            return

        # Make sure to change header/data_header objects to dictionaries
        packet_list = [None] * len(self.raw_data['packet_list'])
        for i, current_packet in enumerate(self.raw_data['packet_list']):
            packet_list[i] = current_packet
            try:
                packet_list[i]['header'] = current_packet['header'].to_dict()
                packet_list[i]['data_header'] = current_packet['data_header'].to_dict()
            except:
                logger.warning(f'header and/or data_header classes '
                               f'cannot be converted and will be removed for packet {current_packet}')
                packet_list[i].pop('header')
                packet_list[i].pop('data_header')

            if self.print:
                print(json.dumps(packet_list[i], indent=2, default=str))

        if self.json_file:
            with open(self.json_file, 'w') as out_json:
                json.dump(packet_list, out_json, indent=2, default=str)

            logger.info(f'{self.json_file} saved')



from libc.stdint cimport int8_t
from libc.stdint cimport uint8_t
from libc.stdint cimport int16_t
from libc.stdint cimport uint16_t
from libc.stdint cimport int32_t
from libc.stdint cimport uint32_t
from .parameter cimport Parameter
from .header cimport TCHeader
from .header cimport TMHeader
from .header cimport Header

cdef extern from "type_conversion.h":
    int8_t to_signed_char(uint8_t x) nogil
    int16_t to_signed_short(uint16_t x) nogil
    int32_t to_signed_long(uint32_t x) nogil
    float to_float(uint32_t x) nogil


cdef class Data:
    cdef bytes ref
    cdef uint8_t* data
    cdef uint32_t size

    cpdef extract_parameter(self, Parameter parameter, uint32_t offset)
    cpdef Header extract_header(self)
    cpdef TMHeader extract_tm_header(self)
    cpdef TCHeader extract_tc_header(self)
    cpdef to_bytes(self)
    cdef int32_t i32(self, uint32_t start, uint32_t offset, uint32_t length) nogil
    cdef uint32_t u32(self, uint32_t start, uint32_t offset, uint32_t length) nogil
    cdef int16_t i16(self, uint32_t start, uint32_t offset, uint32_t length) nogil
    cdef uint16_t u16(self, uint32_t start, uint32_t offset, uint32_t length) nogil
    cdef int8_t i8(self, uint32_t start, uint32_t offset, uint32_t length) nogil
    cdef uint8_t u8(self, uint32_t start, uint32_t offset, uint32_t length) nogil
    cdef tuple time(self, uint32_t start, uint32_t offset, uint32_t length)
    cpdef tuple timep(self, uint32_t start, uint32_t offset, uint32_t length)
    cdef float f32(self, uint32_t start, uint32_t offset, uint32_t length) nogil
    cpdef int32_t i32p(self, uint32_t start, uint32_t offset, uint32_t length)
    cpdef uint32_t u32p(self, uint32_t start, uint32_t offset, uint32_t length)
    cpdef int16_t i16p(self, uint32_t start, uint32_t offset, uint32_t length)
    cpdef uint16_t u16p(self, uint32_t start, uint32_t offset, uint32_t length)
    cpdef int8_t i8p(self, uint32_t start, uint32_t offset, uint32_t length)
    cpdef uint8_t u8p(self, uint32_t start, uint32_t offset, uint32_t length)
    cpdef float f32p(self, uint32_t start, uint32_t offset, uint32_t length)
    cpdef tuple extract_parameters(self, Py_ssize_t offset, list parameters)
    cpdef tuple extract_block(self, Py_ssize_t offset, list parameters, uint32_t counter)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

try:
    from poppy.core.generic.cache import CachedProperty
    from poppy.core.logger import logger
except:
    print('POPPy framework is not installed')

from roc.rpl.constants import TYPE_TO_NUMPY_DTYPE, \
    UNKNOWN_UNIQUE_ID, UNKNOWN_ACK_STATE, UNKNOWN_IDB, \
    RECOVERED_PACKETS, UNKNOWN_STATUS

from roc.rpl.exceptions import RPLError, PacketParsingError
from roc.rpl.packet_structure.data import header_foot_offset


__all__ = ['BasePacket']

def _create_packet_dtype(parameters):
    # track parameter names inside packet, because it is possible to have
    # the same parameter appear multiple times in one packet, so for such
    # parameters, raise an error. Behavior changed compared to previous version
    # where a suffix was added to the parameter name
    tracker = set()

    # create the dtype of the array, taking into account the case where
    # we have block of parameters on which to loop
    dtype = []
    in_block = False
    for parameter in parameters:
        # check we are in block
        in_block = parameter.group > 0

        # store the last parameter if not in block
        if not in_block:
            last_parameter = parameter

        # check if the parameter is already present in the dtype
        if parameter.name in tracker:
            raise RPLError(
                (
                    'The parameter {0} is appearing multiple times in a ' +
                    'packet. It seems that the IDB part of the ROC database ' +
                    'is corrupted.'
                ).format(parameter.name)
            )
        else:
            name = parameter.name

        # create the dtype

        if in_block:
            dtype.append(
                (
                    name,
                    TYPE_TO_NUMPY_DTYPE[parameter.type],
                    parameter.counter if parameter.counter != 0 else 0
                )
            )
        else:
            dtype.append((name, TYPE_TO_NUMPY_DTYPE[parameter.type]))

        # add the parameter to say that it is already used
        tracker.add(parameter.name)

    return dtype

class BasePacket(object):

    def reshape_data(self, packet_list):
        """
        Reshape the packet to match the expected dtype/shape

        - Replace None with NaN-filled/zero-filled numpy array
        - Fill remaining space of parameters block with NaNs/zeroes

        :param packet_list: The content of packets organized as a list of tuples
        :return: The reshaped packets
        :rtype: Iterator[tuple]
        """
        #
        for packet in packet_list:
            if packet is None:
                # create an empty packet
                packet_content = [None] * len(self.data_dtype)
            else:
                # use the packet content
                packet_content = packet

            # loop over packet parameters
            for parameter_position, dtype in enumerate(self.data_dtype):

                if packet is None:
                    logger.warning(f'Empty parameter {dtype[0]} in packet {self.name}')
                    # fill the packet appropriate empty numpy array
                    if len(dtype) > 2:
                        # fill the block with NaN like values
                        packet_content[parameter_position] = np.full(dtype[2], np.nan, dtype=dtype[1])
                    else:
                        # add a single NaN like value with the appropriate dtype
                        packet_content[parameter_position] = np.full(1, np.nan, dtype=dtype[1])

                # if it's a parameter block (multiple parameters with the same dtype)
                if len(dtype) > 2:
                    # compute the difference between the current block length and the expected one
                    nan_block_len = dtype[2] - len(packet_content[parameter_position])

                    # if needed, add NaN to fill the remaining space
                    if nan_block_len > 0:
                        parameter_dtype = packet_content[parameter_position].dtype
                        packet_content[parameter_position].resize(dtype[2])
                        packet_content[parameter_position][-nan_block_len:] = np.full(nan_block_len, np.nan,
                                                                                      dtype=parameter_dtype)

            yield tuple(packet_content)

    @property
    def data(self):
        if len(self._data) > 0:
            # ensure that the packet contents have the expected numpy shape/dtype
            return np.array(list(self.reshape_data(self._data)), dtype=self.data_dtype)
        else:
            # return the empty array
            return np.array(self._data)

    @data.setter
    def data(self, data):
        self._data = data

    def add(self):
        """
        To be called each time a packet of this type is detected, to
        get some information from the data if necessary also.
        """
        self.counter += 1

    def initialize(self):
        """
        Initialize some attributes and data for packets.
        """
        # array for the header common to all type of packets
        header_dtype = self._header_dtype()

        # generate the array for the header
        self.header = np.zeros(self.counter, dtype=header_dtype)
        #self.header.fill(np.nan)

        # generate an array to store utc times
        self.utc_time = np.full(self.counter, np.nan, dtype='datetime64[ns]')

        # generate 2D string array to store TC ack acceptance and execution states
        self.tc_ack_state = np.full([self.counter, 2], '', dtype='U16')

        # Generate string array to store TC unique ID
        self.unique_id = np.full(self.counter, UNKNOWN_UNIQUE_ID, dtype='U24')

        # Generate string array to store TC sequence ID
        self.sequence_name = np.full(self.counter, UNKNOWN_UNIQUE_ID, dtype='U24')

        # generate 2D string array to store idb source/version
        self.idb = np.full([self.counter, 2], UNKNOWN_IDB, dtype='U24')

        # generate string array to store binary
        self.binary = np.full(self.counter, '', dtype=object)

        # generate string array to store sha
        self.sha = np.full(self.counter, '', dtype='U64')

        # Generate 1d-element vector with packet compression status
        self.compressed = np.zeros(self.counter, dtype=int)

        # Generate 1d-element vector for packet status and related message (if any)
        self.status = np.full(self.counter, UNKNOWN_STATUS, dtype=int)
        self.message = np.full(self.counter, '', dtype='U64')

        # if this not a packet with data, create empty array
        if self.skip_data:
            self.data = np.array([])
            return

        # create the dtype for the array of the packet
        self.data_dtype = _create_packet_dtype(self.parameters)

        # generate the array for the data
        self.data = [None, ] * self.counter

    @classmethod
    def initialize_all(cls):
        """
        Initialize all instances of packets for their data.
        """
        for instance in cls.manager.instances:
            instance.initialize()

    def _get_header(self, data):
        # get the header
        return data.extract_header()

    def _get_data_header(self, data):
        # extract the data header
        if self.type == 0:
            data_header = data.extract_tm_header()
        elif self.type == 1:
            data_header = data.extract_tc_header()

        # add it to data
        return data_header

    def _get_data(self, data, header, packet_length):
        # get header and footer offsets
        header_offset, foot_offset = header_foot_offset(
            header.packet_type
        )

        # get values from
        values, offset = data.extract_parameters(
            header_offset,
            self.parameters,
        )

        # add the foot offset
        offset += foot_offset

        # get the size of the packet in bytes
        packet_length = (
            header.packet_length if packet_length is None else packet_length
        )

        # compare computed size from expected size of the header
        computed_size = self.byte_size + offset - 7
        if computed_size != packet_length:
            message = (
                'Packet {0} not aligned.\n Expected {1} ' +
                'found {2} bytes. Check IDB version.'
            ).format(
                self.name,
                packet_length,
                computed_size,
            )

            # If the problem is known, recovers the packet
            if not self.name in RECOVERED_PACKETS:
                logger.debug(f'Recovering {self.name}')
            else:
                # else raise en exception
                raise PacketParsingError(message)

        return values

    def extract_data(self, data,
                     packet_length=None,
                     packet_data=None):
        """
        Given the data of the packet, extract it and store it in the
        right format.
        WARNING - Do not forget to update the compressed version of extract_data
        (roc.rpl.compressed.utils.BaseCompressed.extract_data) if you change this method!)

        :param data: Packet binary source/application data to extract
        :param packet_length: packet length
        :param packet_data: dictionary containing packet metadata
        :return: Extracted packet data parameters
        """
        result = None
        try:
            # extract header
            header = self._get_header(data)
            self.header[self.internal_counter] = header.to_tuple()

            # get the type (TM=0, TC=1) of the packet from the first accession
            self.type = self.header['packet_type'][self.internal_counter]

            # extract data header
            data_header = self._get_data_header(data)
            self.data_header[self.internal_counter] = data_header.to_tuple()

            # do not extract data if empty packet
            if not self.skip_data:
                result = self._get_data(
                    data,
                    header,
                    packet_length,
                )

                # store the values
                self._data[self.internal_counter] = result

                # update the dtype
                for idx, dtype in enumerate(self.data_dtype):
                    if isinstance(result[idx], np.ndarray):
                        # compute the max size of parameter blocks
                        block_len = max(len(result[idx]), dtype[2])

                        # and update the dtype
                        self.data_dtype[idx] = dtype[:2] + (block_len,)
        except Exception as e:
            raise e
        finally:
            if packet_data:
                self.utc_time[self.internal_counter] = np.datetime64(packet_data['utc_time'])
                self.tc_ack_state[self.internal_counter, 0] = packet_data.get('ack_acc_state', UNKNOWN_ACK_STATE)
                self.tc_ack_state[self.internal_counter, 1] = packet_data.get('ack_exe_state', UNKNOWN_ACK_STATE)
                self.unique_id[self.internal_counter] = packet_data.get('unique_id', UNKNOWN_UNIQUE_ID)
                self.sequence_name[self.internal_counter] = packet_data.get('sequence_name', UNKNOWN_UNIQUE_ID)
                self.idb[self.internal_counter, 0] = packet_data.get('idb_source', UNKNOWN_IDB)
                self.idb[self.internal_counter, 1] = packet_data.get('idb_version', UNKNOWN_IDB)
                self.binary[self.internal_counter] = packet_data.get('binary')
                self.sha[self.internal_counter] = packet_data.get('sha')
                self.compressed[self.internal_counter] = packet_data.get('compressed', 0)
                self.status[self.internal_counter] = packet_data.get('status', UNKNOWN_STATUS)
                self.message[self.internal_counter] = packet_data.get('message', '')

            # increment counter of packets (gives the possibility to see a hole
            # inside data if a packet have a problem)
            self.internal_counter += 1

        return result

    @CachedProperty
    def data_header(self):

        # create dtype in function of the category
        if self.type == 0:
            data_header_dtype = self._tm_header_dtype()
        elif self.type == 1:
            data_header_dtype = self._tc_header_dtype()

        # create the array for data field header
        data_header = np.empty(
            self.counter,
            dtype=data_header_dtype,
        )
        #data_header.fill(np.nan)

        # return the array
        return data_header

    @staticmethod
    def _header_dtype():
        """
        Create the dtype of the numpy array associated to the common
        header of packets.
        """
        return [
            ('ccsds_version_number', 'uint8'),
            ('packet_type', 'uint8'),
            ('data_field_header_flag', 'uint8'),
            ('process_id', 'uint8'),
            ('packet_category', 'uint8'),
            ('segmentation_grouping_flag', 'uint8'),
            ('sequence_cnt', 'uint16'),
            ('packet_length', 'uint16'),
        ]

    @staticmethod
    def _tm_header_dtype():
        return [
            ('spare_1', 'uint8'),
            ('pus_version', 'uint8'),
            ('spare_2', 'uint8'),
            ('service_type', 'uint8'),
            ('service_subtype', 'uint8'),
            ('destination_id', 'uint8'),
            ('time', TYPE_TO_NUMPY_DTYPE['time']),
        ]

    @staticmethod
    def _tc_header_dtype():
        return [
            ('ccsds_secondary_header_flag', 'uint8'),
            ('pus_version', 'uint8'),
            ('ack_execution_completion', 'uint8'),
            ('ack_execution_progress', 'uint8'),
            ('ack_execution_start', 'uint8'),
            ('ack_acceptance', 'uint8'),
            ('service_type', 'uint8'),
            ('service_subtype', 'uint8'),
            ('source_id', 'uint8'),
        ]

    def __repr__(self):
        return 'Packet[{0}]'.format(self.name)



from libc.stdint cimport uint32_t

cdef class Parameter:
    cdef public uint32_t byte
    cdef public uint32_t bit
    cdef public uint32_t group
    cdef public uint32_t block_size
    cdef public uint32_t counter
    cdef public uint32_t length
    cdef public double maximum
    cdef public str name
    cdef public str type



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""PacketParser time module"""
import os
import os.path as osp
from datetime import datetime

import numpy as np
from poppy.core.configuration import Configuration
from poppy.core.generic.metaclasses import Singleton
from poppy.core.logger import logger

from roc.rpl.constants import SOLAR_ORBITER_NAIF_ID
from roc.rpl.packet_structure.data import Data
from roc.rpl.time.spice import SpiceHarvester


__all__ = ['Time']


# SEC TO NANOSEC conversion factor
SEC_TO_NANOSEC = np.timedelta64(1000000000, 'ns')

# J2000 base time
J2000_BASETIME = np.datetime64('2000-01-01T12:00', 'ns')

# Number of leap nanoseconds at J2000 origin (32 sec)
J2000_LEAP_NSEC = np.timedelta64(32000000000, 'ns')

# RPW CCSDS CUC base time
RPW_BASETIME = np.datetime64('2000-01-01T00:00', 'ns')

# Delta nanosec betwen TAI and TT time bases (TAI = TT + 32.184 sec)
DELTA_NSEC_TAI_TT = np.timedelta64(32184000000, 'ns')

# Delta nanosec between RPW and J2000 epoch (J2000 = RPW + 12h)
DELTA_NSEC_RPW_J2000 = RPW_BASETIME - J2000_BASETIME

# RPW BASETIME in TT2000 (nanoseconds since J2000)
TT2000_RPW_BASETIME = DELTA_NSEC_TAI_TT + \
    J2000_LEAP_NSEC + DELTA_NSEC_RPW_J2000

# TT2000 BASETIME in Julian days
TT2000_JD_BASETIME = 2451545.0

# CUC fine part max value
CUC_FINE_MAX = 65536

# NASA CDF leapsecond table filename
CDF_LEAPSECONDSTABLE = 'CDFLeapSeconds.txt'


class TimeError(Exception):
    """Errors for the time."""


class Time(object, metaclass=Singleton):
    """
    PacketParser Time class.
    (Singleton: only one instance can be called.)

    """

    def __init__(self,
                 kernel_date=None,
                 predictive=True,
                 flown=True,
                 spice_kernel_path='',
                 lstable_file=None,
                 no_spice=False):

        # Set SPICE attributes
        self.kernel_date = kernel_date
        self.predictive = predictive
        self.flown = flown
        self.kernel_path = spice_kernel_path
        self.no_spice = no_spice

        # Initialize cached property and unloaded kernels (if any)
        self.reset()

        # If no spice, need a CDF leap second table file
        if no_spice and lstable_file:
            self.leapsec_file = lstable_file

    def reset(self, spice=True, lstable=True):
        """
        Reset spice and lstable related attribute values of the Time object.

        :param spice: If True, reset spice attribute
        :param lstable: If True, reset lstable attribute
        :return:
        """

        if lstable:
            self._leapsec_file = ''
            self._lstable = None
        if spice:
            # Make sure to unload all kernels
            if hasattr(self, '_spice') and self._spice:
                self._spice.unload_all()

            self.kernel_path = ''
            self._spice = None

    @property
    def leapsec_file(self):
        return self._leapsec_file

    @leapsec_file.setter
    def leapsec_file(self, lstable_file):
        """
        Set input leap seconds table file (and reload table if necessary).

        :param lstable_file: Path to the leap seconds table file, as provided in the NASA CDF software
        :return:
        """

        if lstable_file and lstable_file == self._leapsec_file:
            logger.debug(f'{lstable_file} already loaded')
        else:

            from maser.tools.time import Lstable

            if not lstable_file:
                if 'pipeline.environment.CDF_LEAPSECONDSTABLE' in Configuration.manager:
                    lstable_file = Configuration.manager['pipeline'][
                        'environment.CDF_LEAPSECONDSTABLE']
                elif 'CDF_LEAPSECONDSTABLE' in os.environ:
                    lstable_file = os.environ['CDF_LEAPSECONDSTABLE']
                elif 'CDF_LIB' in os.environ:
                    lstable_file = osp.join(
                        os.environ['CDF_LIB'], '..', CDF_LEAPSECONDSTABLE)
                else:
                    lstable_file = CDF_LEAPSECONDSTABLE

            try:
                lstable = Lstable(file=lstable_file)
            except:
                logger.error(f'New leap second table "{lstable_file}" cannot be loaded!')
            else:
                self._lstable = lstable
                self._leapsec_file = lstable_file

    @property
    def lstable(self):
        if not self._lstable:
            # If not alread loaded,
            # get lstable (Set self.leapsec_file to None will trigger loading table file from config. file or
            # environment variables)
            self.leapsec_file = None

        return self._lstable

    @property
    def spice(self):
        if not self._spice and not self.no_spice:
            self.spice = self._load_spice()
        return self._spice

    @spice.setter
    def spice(self, value):
        self._spice = value

    def _load_spice(self):
        """
        Load NAIF Solar Orbiter SPICE kernels for the current session.

        :return:
        """

        if not self.kernel_path:
            self.kernel_path = SpiceHarvester.spice_kernel_path()

        # Load SPICE kernels (only meta kernels for the moment)
        return SpiceHarvester.load_spice_kernels(self.kernel_path,
                                                 only_mk=True,
                                                 by_date=self.kernel_date,
                                                 predictive=self.predictive,
                                                 flown=self.flown)

    def obt_to_utc(self, cuc_time,
                   to_datetime=False,
                   to_tt2000=False):
        """
        Convert input RPW CCSDS CUC time(s) into UTC time(s).

        :param cuc_time: RPW CCSDS CUC time(s)
        :param to_datetime: If True, return UTC time as datetime object
        :param to_tt2000: If True, return TT2000 epoch time (nanosec since J2000) instead of UTC (data type is int)
        :param: predictive: If True, then use predictive SPICE kernels instead of as-flown
        :param: kernel_date: datetime object containing the date for which SPICE kernels must be loaded (if not passed, then load the latest kernels)
        :param: reset: If True, then reset Time instance harvester
        :param: no_spice: If True, dot not use SPICE toolkit to compute UTC time
        :return: utc time(s) as numpy.datetime64 datatype (datetime object if to_datetime is True)
        """

        # Make sure that cuc_time is a numpy array
        cuc_time = np.array(cuc_time)

        # Check that input cuc_time has the good shape
        shape = cuc_time.shape
        if shape == (2,):
            cuc_time = cuc_time.reshape([1, 2])

        # If no_spice = True ...
        if self.no_spice:
            # If SPICE not available, use internal routines
            # (assuming no drift of OBT or light time travel corrections
            # . It should be used for on-ground test only)
            utc_time = self.cuc_to_utc(cuc_time, to_tt2000=to_tt2000)
        else:
            spice = self.spice

            nrec = len(cuc_time)

            if not to_tt2000:
                utc_time = np.empty(nrec, dtype='datetime64[ns]')
            else:
                utc_time = np.empty(nrec, dtype=int)

            # Then use SpiceManager to compute utc
            for i, current_time in enumerate(cuc_time):

                obt = spice.cuc2obt(current_time)
                et = spice.obt2et(SOLAR_ORBITER_NAIF_ID, obt)

                if not to_tt2000:
                    utc_time[i] = spice.et2utc(et)
                else:
                    tdt = spice.et2tdt(et)
                    utc_time[i] = tdt * SEC_TO_NANOSEC

        # Convert to datetime object if to_datetime keyword is True
        if not to_tt2000 and to_datetime:
            utc_time = Time.datetime64_to_datetime(utc_time)

        return utc_time

    @staticmethod
    def cuc_to_nanosec(cuc_time, j2000=False):
        """
        Given a CCSDS CUC coarse and fine cuc_time in the numpy array format,
        returns it in nanoseconds since the defined epoch.

        :param cuc_time: input RPW CCSDS CUC time provided as numpy array
        :return: Return time in nano seconds since RPW origin (or J2000 if j2000 keyword is True) as numpy.uint64 array
        """
        # first convert cuc time into scet
        scet = Time.cuc_to_scet(cuc_time)

        # Then convert scet in nanosec
        nanosec = scet * np.float64(SEC_TO_NANOSEC)

        if j2000:
            nanosec -= np.float64(DELTA_NSEC_RPW_J2000)

        return nanosec.astype(np.uint64)

    @staticmethod
    def cuc_to_scet(cuc_time):
        """
        Return input RPW CCSDS CUC time as spacecraft elapsed time since RPW origin

        :param cuc_time: input RPW CCSDS CUC time as numpy array
        :return: spacecraft elapsed time since RPW origin as numpy.float46 array
        """

        # Make sure that input cuc_time is a numpy array
        if not isinstance(cuc_time, np.ndarray):
            cuc_time = np.array(cuc_time, dtype=np.float64)

        # Check that input cuc_time has the good shape
        shape = cuc_time.shape
        if shape == (2,):
            cuc_time = cuc_time.reshape([1, 2])

        # convert coarse part into nanoseconds
        coarse = cuc_time[:, 0].astype('float64')

        # same for fine part
        fine = cuc_time[:, 1].astype('float64')
        fine = fine / np.float64(CUC_FINE_MAX)

        return coarse + fine

    def cuc_to_utc(self, cuc_time,
                   from_nanosec=False,
                   to_tt2000=False):
        """
        Convert RPW CCSDS CUC time(s) into in UTC time(s)

        :param cuc_time: numpy uint array of RPW CCSDS CUC time(s)
        :param from_nanosec: If True, then input RPW CCSDS CUC time array is expected in nanoseconds
        :param to_tt2000: Convert input CCSDS CUC time into TT2000 epoch instead of UTC
        :return: numpy datetime64 array of UTC time(s)
        """

        if not from_nanosec:
            # Convert cuc to nanosec since RPW_BASETIME
            cuc_nanosec = self.cuc_to_nanosec(cuc_time)
        else:
            cuc_nanosec = cuc_time.astype(np.unint64)

        # Convert to TT20000 epoch time
        tt2000_epoch = cuc_nanosec + TT2000_RPW_BASETIME

        if to_tt2000:
            return tt2000_epoch
        else:
            return self.tt2000_to_utc(tt2000_epoch)

    def tt2000_to_utc(self, tt2000_epoch,
                      leap_sec=None,
                      to_datetime=True):
        """
        Convert input TT2000 epoch time into UTC time

        :param tt2000_epoch: TT2000 epoch time to convert
        :param leap_sec: number of leap second to apply
        :param to_datetime: return utc time as a datetime object (microsecond resolution)
        :return: UTC time (datetime.datetime if to_datetime=True, format returned by spice.tdt2uct otherwise)
        """

        if self.no_spice:
            # TT2000 to TT time
            tt_time = tt2000_epoch + J2000_BASETIME

            # Get leap seconds in nanosec
            if not leap_sec:
                lstable = self.lstable

                leap_nsec = np.array([lstable.get_leapsec(Time.datetime64_to_datetime(tt)) * SEC_TO_NANOSEC
                                      for tt in tt_time], dtype='timedelta64[ns]')
            else:
                leap_nsec = leap_sec * np.uint64(SEC_TO_NANOSEC)

            utc_time = tt_time - DELTA_NSEC_TAI_TT - leap_nsec
        else:
            spice = self.spice
            # Convert input TT2000 time in seconds (float)
            tdt = float(tt2000_epoch) / float(SEC_TO_NANOSEC)
            utc_time = spice.tdt2utc(tdt)
            if to_datetime:
                utc_time = datetime.strptime(utc_time,
                                             spice.TIME_ISOC_STRFORMAT)

        return utc_time

    def utc_to_tt2000(self, utc_time,
                      leap_sec=None):
        """
        Convert an input UTC time value into a TT2000 epoch time

        :param utc_time: an input UTC time value to convert (numpy.datetime64 or datetime.datetime type)
        :param leap_sec: number of leap seconds at the utc_time (only used if no_spice=True)
        :param no_spice: If True, do not use SPICE toolkit to compute tt2000 time.
        :return: TT2000 epoch time value (integer type)
        """

        # Only works with scalar value for input utc_time
        if isinstance(utc_time, list):
            utc_time = utc_time[0]

        if isinstance(utc_time, datetime):
            utc_dt64 = Time.datetime_to_datetime64(utc_time)
            utc_dt = utc_time
        elif isinstance(utc_time, np.datetime64):
            utc_dt64 = utc_time
            utc_dt = Time.datetime64_to_datetime(utc_time)
        else:
            logger.error('Unknown type for input utc_time!')
            return None

        # Compute without SPICE
        if self.no_spice:
            # Get leap seconds in nanosec
            if not leap_sec:
                lstable = self.lstable
                leap_nsec = lstable.get_leapsec(utc_dt) * SEC_TO_NANOSEC
            else:
                leap_nsec = leap_sec * np.uint64(SEC_TO_NANOSEC)

            # Get nanoseconds since J2000
            # And add leap seconds + the delta time between TAI and TT
            tt2000_epoch = int(utc_dt64 - J2000_BASETIME) \
                + DELTA_NSEC_TAI_TT + np.timedelta64(leap_nsec, 'ns')
        else:
            # Use SPICE toolkit
            spice = self.spice

            # Convert input utc to string
            utc_string = spice.utc_datetime_to_str(utc_dt)
            # Convert utc string into ephemeris time then tdt
            et = spice.utc2et(utc_string)
            tdt = spice.et2tdt(et)

            # Get tt2000_epoch in nanoseconds
            tt2000_epoch = int(tdt * SEC_TO_NANOSEC)

        return tt2000_epoch

    @staticmethod
    def cuc_to_microsec(time):
        """
        cuc_to_microsec.

        Convert a RPW CUC format time into microsec.
        """
        # convert coarse part into microseconds
        coarse = time[:, 0].astype('uint64') * 1000000

        # same for fine part
        fine = np.rint((time[:, 1].astype('uint64') * 1000000) / 65536).astype(
            'uint64'
        )

        return coarse + fine

    @staticmethod
    def cuc_to_numpy_datetime64(time):
        """
        Given a CUC format RPW time in the numpy array format
        , returns it np datetime64 format.
        (For more details about RPW CUC format see RPW-SYS-SSS-00013-LES)

        :param time: array of input CCSDS CUC times (coarse, fine)
        :return: datetime64 object of input CUC time
        """

        ns = Time.cuc_to_nanosec(time)
        base_time = RPW_BASETIME

        dt64 = np.array([
                        base_time + np.timedelta64(int(nsec), 'ns')
                        for nsec in ns
                        ], dtype=np.datetime64)

        #dt64 = Time.datetime64_to_datetime(dt64)

        return dt64

    @staticmethod
    def cuc_to_datetime(time):
        """
        Convert input CCSDS CUC RPW time into datetime object.

        :param time: array of input CCSDS CUC times (coarse, fine)
        :return: CUC time as datetime
        """

        dt = Time.cuc_to_numpy_datetime64(time)
        return Time.datetime64_to_datetime(dt)

    @staticmethod
    def numpy_datetime64_to_cuc(time, synchro_flag=False):
        """
        Given a numpy.datetime64, returns CCSDS CUC format time in the numpy array format.
        (For more details about RPW CUC format see RPW-SYS-SSS-00013-LES)

        :param time: numpy.datetime64 to convert to CCSDS CUC
        :param synchro_flag: If True, time synchro flag bit is 1 (=not sync), 0 (=sync) otherwise
        :return:numpy.array with [CUC coarse, fine, sync_flag]
        """

        # convert time into nanoseconds since 2000-01-01T00:00:00
        base_time = RPW_BASETIME

        nanoseconds = np.timedelta64(time - base_time, 'ns').item()

        # compute coarse part
        coarse = nanoseconds // SEC_TO_NANOSEC

        # compute fine part
        fine = np.rint(((nanoseconds - coarse * SEC_TO_NANOSEC) /
                        SEC_TO_NANOSEC) * 65536).astype('uint16')

        return np.array([[coarse, fine, int(synchro_flag)]])

    @staticmethod
    def extract_cuc(cuc_bytes):
        """
        Extract RPW CCSDS CUC time 48 bytes and return
        coarse, fine and sync flag

        :param cuc_bytes: 48-bytes CUC time
        :return: coarse, fine and sync flag
        """

        if not isinstance(cuc_bytes, bytes):
            cuc_bytes = bytes(cuc_bytes)

        nbytes = len(cuc_bytes)
        if nbytes != 48:
            logger.error(f'Expect 48 bytes, but {nbytes} found for input CUC time!')
            return None, None, None

        return Data(cuc_bytes, nbytes).timep(0, 0, 48)

    @staticmethod
    def str2datetime64(string):
        """
        Convert an input string in numpy.datetime64

        :param string: string of ISO8601 format 'YYYY-MM-DDThh:mm:ss.ffffffZ'
        :return: numpy.datetime64 time
        """
        return np.datetime64(string)

    @staticmethod
    def datetime64_to_datetime(datetime64_input):
        """
        Convert numpy.datetime64 object into datetime.datetime object

        :param datetime64_input: input numpy.datetime64 object to convert
        :return: datetime.datetime object
        """
        return datetime64_input.astype('M8[us]').astype('O')

    @staticmethod
    def datetime_to_datetime64(datetime_input):
        """
        Convert datetime.datetime object into numpy.datetime64 object

        :param datetime_input: datetime.datetime object to convert
        :return: numpy.datetime64 object
        """
        return np.datetime64(datetime_input)

    @staticmethod
    def tt2000_to_jd(tt2000_time):
        """
        Convert input TT2000 format time into Julian days

        :param tt2000_time: TT2000 format time (i.e., nanoseconds since J2000)
        :return: time in julian days (double)
        """
        return (float(tt2000_time) / (1000000000.0 * 3600.0 * 24.0)) + TT2000_JD_BASETIME

    @staticmethod
    def jd_to_tt2000(jd_time):
        """
        Convert input Julian day into TT2000 time

        :param jd_time: Julian day time
        :return: time in TT2000 format time (i.e., nanoseconds since J2000)
        """
        return (float(jd_time) - TT2000_JD_BASETIME) * (1000000000.0 * 3600.0 * 24.0)

    def jd_to_datetime(self, jd_time):
        """
        Convert input Julian day into datetime.datetime object

        :param jd_time: Julian day time
        :return: datetime.datime object
        """
        # convert to UTC first
        utc_time = self.tt2000_to_utc(
            np.array([
                Time.jd_to_tt2000(jd_time)]), to_datetime=True)
        return utc_time




from roc.rpl.constants import INVALID_PACKET_DATA, VALID_PACKET, INVALID_PACKET_HEADER
from roc.rpl.exceptions import PacketParsingError
from roc.rpl.packet_structure.data import Data

from roc.rpl.packet_parser import PacketParser
from roc.rpl.packet_parser.parser.identify_packets import identify_packets

__all__ = ["parse_packets"]

cpdef dict parse_packet(int packet_idx, dict packet_data, object packet_parser):
    cdef:
        object name
        object srdb_id
        object palisade_id
        object raw_data
        object utc_time
        object byte_data
        object packet
        object result
        object message
        object status
        object comment

    # now extract parameter values for each identified packet

    # name of the packet
    name = packet_data.get('srdb_id')
    palisade_id = packet_data.get('palisade_id')
    utc_time = packet_data.get('utc_time')
    raw_data = packet_data.get('binary')
    status = packet_data.get('status')
    comment = packet_data.get('comment')
    if comment is None:
        packet_data['comment'] = ''

    # If not valid packet skip data extraction
    if (name is None or
            palisade_id is None or
            raw_data is None or
            status != VALID_PACKET):
        packet_data['comment'] += f'Input packet cannot be parsed ({packet_data}). '
        return packet_data

    # logger.debug(f'Extracting data for packet {palisade_id} {{{name}}} at
    # {utc_time}...')

    # convert the hexa raw data in bytes
    byte_data = PacketParser.hex_to_bytes(raw_data)

    # only put in the database events with data inside
    if byte_data is None:
        packet_data['comment'] += "Packet (#{0} {1}, {2}, {3}) doesn't have raw data, skip data extraction! ".format(
            packet_idx, palisade_id, name, utc_time)
        return packet_data

    # check that the packet is a correct packet name
    if palisade_id not in packet_parser.idb.packets:
        packet_data['comment'] += ("Packet (#{0}, {1}, {2} {3}) has not been "
                                   "recognized in the current IDB, skip data extraction! "
                                   ).format(packet_idx, palisade_id, name, utc_time)
        packet_data['status'] = INVALID_PACKET_HEADER
        return packet_data

    # get the packet and extract the data
    packet = packet_parser.packet(palisade_id)
    try:
        result = packet.extract_data(
            Data(bytes(byte_data), len(byte_data)),
            packet_data=packet_data,
        )
        packet_data['data'] = result
    # cases for packet parsing
    except PacketParsingError as e:
        packet_data["status"] = INVALID_PACKET_DATA
        packet_data['comment'] += e.message + '. '
        packet.is_valid = False
        packet_parser.extract_error(e.message)
    # system exit
    except SystemExit:
        raise
    # keyboard interrupt
    except KeyboardInterrupt:
        raise
    # all other cases (compression, etc)
    except Exception as e:
        packet_data["status"] = INVALID_PACKET_DATA
        message = (f"Issue with packet "
                   f"(#{packet_idx} {palisade_id} {{{name}}} "
                   f"at {utc_time}): {e}. ")
        packet_data["comment"] += message
        packet.is_valid = False
        packet_parser.extract_error(message)

    return packet_data

cpdef list parse_packets(object packet_parser, list packet_list,
                         start_time=None,
                         end_time=None,
                         no_identification=False):
    """
    Parse input RPW TM/TC packets and store their information
    in the packet structure.

    :param packet_parser: instance of packet_parser object
    :param packet_list: List of raw binary data of the packets
    :return: Store results in the PacketParser object
    """
    cdef:
        list identified_packet_list
        list parsed_packet_list
        Py_ssize_t packet_idx
        long packet_count
        dict current_packet

    # Extract packet header info and names
    if not no_identification:
        identified_packet_list = identify_packets(
            packet_parser, packet_list,
            start_time=start_time,
            end_time=end_time)
    else:
        identified_packet_list = packet_list

    # Add packets to the Packet class instance
    added_list = [packet_parser.add_packet(current_packet)
                  for current_packet in identified_packet_list]

    # If only not valid packet, exit method
    packet_count = len(identified_packet_list)
    if packet_count == 0:
        return identified_packet_list

    # initialize the packet structure
    packet_parser.packet.initialize_all()

    parsed_packet_list = [parse_packet(packet_idx,
                                       identified_packet_list[packet_idx],
                                       packet_parser)
                          for packet_idx in range(packet_count)]

    return parsed_packet_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
from datetime import datetime

from poppy.core.conf import settings
from poppy.core.logger import logger

__all__ = ['TYPE_TO_NUMPY_DTYPE',
           'VALID_PACKET',
           'INVALID_PACKET_HEADER',
           'INVALID_PACKET_DATA',
           'OUTSIDE_RANGE_PACKET',
           'INVALID_PACKET_CDATA',
           'INVALID_PACKET_TIME',
           'UNKNOWN_STATUS',
           'TC_PACKET_TYPE', 'TM_PACKET_TYPE',
           'UNKNOWN_PACKET', 'TIME_ISO_STRFORMAT',
           'TIME_DAILY_STRFORMAT', 'TIME_RANGE_STRFORMAT',
           'TCREPORT_STRTFORMAT', 'INVALID_UTC_DATETIME',
           'INVALID_UTC_TIME', 'UNKNOWN_ACK_STATE',
           'UNKNOWN_UNIQUE_ID', 'TM_RAW_DATA_DEF',
           'RAW_DATA_DEF', 'TC_RAW_DATA_DEF', 'UNKNOWN_IDB',
           'SPICE_KERNEL_PATTERN', 'TC_ACK_STATE_PASSED',
           'SOLAR_ORBITER_NAIF_ID', 'KERNEL_CATEGORY',
           'TC_ACK_STATE_PASSED', 'EXCLUDED_PACKETS',
           'RECOVERED_PACKETS', 'SCOS_HEADER_BYTES',
           'PIPELINE_DATABASE', 'TEST_DATABASE',
           ]

TYPE_TO_NUMPY_DTYPE = {
    'uint16': 'uint16',
    'uint8': 'uint8',
    'uint32': 'uint32',
    'int16': 'int16',
    'int8': 'int8',
    'int32': 'int32',
    'float32': 'float32',
    'time': ('uint32', 3)
}

# Packet status possible flags
VALID_PACKET = 0
INVALID_PACKET_HEADER = 1
INVALID_PACKET_DATA = 2
OUTSIDE_RANGE_PACKET = 3
# same than INVALID_PACKET_DATA but for invalid compressed data
INVALID_PACKET_CDATA = 4
INVALID_PACKET_TIME = 5
UNKNOWN_STATUS = -1

# TM/TC packet_type CCSDS ID
TM_PACKET_TYPE = 0
TC_PACKET_TYPE = 1

# Unknown Packet flag
UNKNOWN_PACKET= 'UNKNOWN'

# TC ack. state flag
TC_ACK_STATE_PASSED = 'PASSED'

# Datetime string formats
INPUT_DATETIME_STRFTIME = '%Y-%m-%dT%H:%M:%S'
TIME_ISO_STRFORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
TCREPORT_STRTFORMAT = TIME_ISO_STRFORMAT
TIME_DAILY_STRFORMAT = '%Y%m%d'
TIME_RANGE_STRFORMAT = '%Y%m%dT%H%M%S'

INVALID_UTC_TIME = '2000-01-01T00:00:00.000000Z'

UNKNOWN_UNIQUE_ID = 'UNKNOWN'
UNKNOWN_ACK_STATE = 'UNKNOWN'
UNKNOWN_IDB = 'UNKNOWN'

# Convert invalid UTC time string into datetime
INVALID_UTC_DATETIME = datetime.strptime(INVALID_UTC_TIME,
                                        TCREPORT_STRTFORMAT)

# Default values for RawData packet list dictionary
RAW_DATA_DEF = {
    'srdb_id': None,
    'palisade_id': None,
    'status': VALID_PACKET,
    'utc_time': INVALID_UTC_DATETIME,
    'binary': None,
    'type': None,
}

# Default values for RawData TM packet list dictionary
TM_RAW_DATA_DEF = {
    'sync_flag': None,
}

# Default values for RawData TC packet list dictionary
TC_RAW_DATA_DEF = {
    'ack_acc_state': None,
    'ack_exe_state': None,
}

# Special packets
EXCLUDED_PACKETS = [
                    'RACK_SUCCESS_EXE_COMMAND',  # MEB GSE packet
                    'RACK_HK', # MEB GSE packet
                    ]

# Packets, known to be inconsistent but must be parsed anyway
RECOVERED_PACKETS = [
                    'TM_DPU_RWF_HK', # 102 bytes packet size found, but 101 bytes expected
                    ]



# Give for each kernel type the glob file pattern,
# The field(s) to extract as regex and the number of field(s) to extract
SPICE_KERNEL_PATTERN = {'ck': ['solo_ANC_soc-*_????????-????????_V??.bc',
                               '\_([0-9]{8})\-([0-9]{8})\_V([0-9]{2}).bc', 3],
                        'fk':['solo_ANC_soc-*_V??.tf',
                              '\_V([0-9]{2}).tf', 1],
                        'ik':['solo_ANC_soc-*_V??.ti',
                              '\_V([0-9]{2}).ti', 1],
                        'mk':['solo_ANC_soc-*.tm',
                              '\_V([0-9]{3})\_([0-9]{8})\_([0-9]{3}).tm', 3],
                        'sclk': ['solo_ANC_soc-*_????????_????????_V??.tsc',
                                 '\_([0-9]{8})\_([0-9]{8})\_V([0-9]{2}).tsc', 3],
                        'spk':['solo_ANC_soc-*_????????-????????_V??.bsp',
                                '\_([0-9]{8})\-([0-9]{8})\_V([0-9]{2}).bsp', 3],
                        'lsk':['*.tls',
                               '([0-9]{4}).tls', 3],
                        'pck':['*.*', None, 0],
                        }



# NAIF SPICE ID of Solar Orbiter
SOLAR_ORBITER_NAIF_ID = -144

# Predictive and as-flown kernel category
# predictive should be loaded first
KERNEL_CATEGORY = ['pred', 'flown']

# DDS TmRaw SCOS header length in bytes
SCOS_HEADER_BYTES = 76

# Load pipeline database identifier
try:
    PIPELINE_DATABASE = settings.PIPELINE_DATABASE
except:
    PIPELINE_DATABASE = 'PIPELINE_DATABASE'
    logger.warning(f'settings.PIPELINE_DATABASE not defined for {__file__}, \
                     use "{PIPELINE_DATABASE}" by default!')

try:
    TEST_DATABASE = settings.TEST_DATABASE
except:
    TEST_DATABASE = 'TEST_DATABASE'
    logger.warning(f'settings.TEST_DATABASE not defined for {__file__}, \
                     use "{TEST_DATABASE}" by default!')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

from roc.rpl.constants import INVALID_PACKET_CDATA
from roc.rpl.exceptions import RPLError
from roc.rpl.rice.rice import Compressor
from roc.rpl.packet_structure.data import header_foot_offset
from roc.rpl.packet_structure.data import Data

__all__ = ['find', 'find_index', 'get_base_packet']


def get_base_packet(Packet):
    """
    Generate the base class to use for compressed packets in the case of TDS.
    """
    # check if the class is already defined
    # base class for the compressed packets of LFR
    class BaseCompressed(Packet):
        """
        Base class for the compressed packets of LFR. Define the special
        behaviour of the treatment for those packets, that should mimic the non
        compressed (NC) ones.
        """
        # create an object allowing to do decompression of data
        compressor = Compressor()

        def __init__(self, name):
            # init the packet as the other
            super(BaseCompressed, self).__init__(name)

            # store the name of the non compressed packet
            if self.name.endswith('_C'):
                self.nc_name = self.name[:-2]
            else:
                # raise exception because the packet is not handled correctly
                raise RPLError(
                    (
                        'Trying to register {0} as compressed packet while ' +
                        'not following the name pattern'
                    ).format(self.name)
                )

            # get the associated NC packet
            self.nc_packet = Packet(self.nc_name)

            # and parameter marking start of the data block for the packet
            self.idx_data_start, self.data_start = find_index(
                self.parameters,
                self.DATA_START,
            )

            # same for the packet indicating the number of compressed packet
            self.idx_data_count, self.data_count = find_index(
                self.parameters,
                self.DATA_COUNT,
            )

            # same for the divergence with the NC packet
            self.idx_start, self.start = find_index(
                self.parameters,
                self.START,
            )

            # same for the NC packet
            self.nc_start = find(self.nc_packet.parameters, self.NC_DATA_START)

            # the index of the parameter with the number of blocks
            self.idx_nblocks, self.nblocks = find_index(
                self.parameters,
                self.NBLOCK,
            )

            # get the block size in byte to be able to know the size of the
            # data to decompress
            self.block_size = self.nc_start.block_size

            # compute the header offset (with know that it is TM packet)
            self.category = 0
            self.header_offset, _ = header_foot_offset(self.category)

            # the starting position for the compressed data
            self.start_position = self.data_start.byte + self.header_offset

            # no data reorganization by default
            self.is_compressed = True

            # valid packet flag
            self.is_valid = True


        def add(self):
            """
            Increment the number of packets of the NC packet because data will
            be added inside it.
            """
            self.nc_packet.add()

        def extract_data(self, data,
                         packet_length=None,
                         packet_data=None):
            """
            Extract the data as it was the NC packet. The hack is to modify the
            data buffer to reflect the structure of the NC packet.
            """
            try:
                # start by extracting the parameters until the divergence
                # between compressed and non compressed packets
                values, _ = data.extract_parameters(
                    self.header_offset,
                    self.parameters[:self.idx_data_start],
                )

                # get the compressed data part of the packet
                compressed_data = self.compressed_data(data, values)

                # get the number of bytes to uncompress
                size = self.uncompressed_byte_size(data, values)

                # get the uncompressed data by the method specified for each
                # instrument
                self.uncompressed = self.uncompressed_data(
                    compressed_data,
                    values,
                    size,
                )

                # check that the expected size for uncompressed data and the
                # extracted one from the compressed packet are the same
                if size != len(self.uncompressed):
                    raise RPLError(
                        (
                            "The uncompressed data for packet {0} doesn't" +
                            'have the same size as the expected one.\n' +
                            'Expected: {1} bytes\t Extracted: {2} bytes'
                        ).format(
                            self.name,
                            size,
                            len(self.uncompressed),
                        )
                    )

                # create the new data from the information
                new_data = self.data_substitution(data, self.uncompressed)

                # Set compressed status to 1
                packet_data['compressed'] = 1

                # extract the data as it is the NC packet
                self.nc_packet.extract_data(
                    new_data,
                    packet_length=new_data.data_size - 7,
                    packet_data=packet_data,
                )
            # If an exception is raised, set packet status to INVALID_PACKET_CDATA
            # and save error message
            except Exception as e:
                self.is_valid = False
                packet_data['status'] = INVALID_PACKET_CDATA
                packet_data['comment'] = e.message
            # gives the possibility to see a hole in data if something went
            # wrong
            finally:
                self.counter += 1

        def compressed_data(self, data, values):
            """
            Return the compressed data part of a compressed packet.
            """
            # get the size of compressed data and compare to the size
            # obtained from the position
            compressed_size = self.compressed_byte_size(values)

            # get the compressed data
            buf = data.to_bytes()[self.start_position:]

            # check the size of the data to what is expected from the
            # decommuted values of the compressed packet
            if compressed_size != len(buf):
                raise RPLError(
                    (
                        'The compressed data size is not coherent in the' +
                        ' packet parameters and from the bytes ' +
                        'themselves.\nFrom packet parameters: {0} bytes' +
                        '\nFrom data size: {1}'
                    ).format(
                        compressed_size,
                        len(buf),
                    )
                )

            # return the compressed data
            return Data(buf, len(buf))

        def uncompressed_data(self, compressed_data, values, size):
            """
            Return the uncompressed data from the data of the packet and the
            decommuted values of the compressed packet. Packets for LFR are
            special, because compressed data is organized by blocks of values.
            So we need to decompressed each block, then reorganize data
            according to the good structure as for the non compressed packet.
            """
            # initialize the offset when going though the compressed packet
            offset = 0

            # initialize the data buffer that will need to be transposed at the
            # end
            data = bytes()

            # compute the size in bytes of the uncompressed block
            uncompressed_block_size = size // self.nc_start.group

            # loop over the number of blocks of compressed data. This number is
            # directly the number of parameters inside a block of data for the
            # non compressed packet.
            for block in range(self.nc_start.group):
                # read the size of the block (16 bits)
                block_size = compressed_data.u16p(offset, 0, 16)

                # increment offset
                offset += 2

                # create a buffer for the compressed data with the good size
                buf = Data(
                    compressed_data.to_bytes()[offset:offset + block_size],
                    block_size,
                )

                # uncompress the data (and truncate to the size of expected
                # data because apparently the returned size can be superior to
                # the wanted size)
                data += self.compressor.uncompress(
                    buf,
                    uncompressed_block_size,
                ).to_bytes()[:uncompressed_block_size]

                # add the size of the compressed data
                offset += block_size

            # convert the concatenated byte array into a numpy array to be able
            # to transpose it and reorder correctly the data by sample
            data = np.frombuffer(data, dtype='uint8').view('uint16')
            data.shape = (
                self.nc_start.group,
                len(data) // self.nc_start.group
            )

            # transpose the data if a reorganization is needed and convert to
            # bytes
            return data.T.byteswap().tobytes()

        def data_substitution(self, data, uncompressed):
            """
            Create a new data bytes array in order to simulate an non
            compressed packet with the data from an uncompressed packet,
            allowing to decommute and store a compressed packet as an
            uncompressed one.
            """
            new_bytes = data.to_bytes()[:self.start.byte + self.header_offset]
            new_bytes += uncompressed
            new_data = Data(new_bytes, len(new_bytes))
            return new_data

        def compressed_byte_size(self, values):
            """
            Given the decommuted values of the packet before the data part,
            returns the size in bytes of the compressed data, as expected by
            the packet.
            """
            # check the maximal value of the block parameter
            if values[self.idx_data_count] > self.data_count.maximum:
                raise RPLError(
                    'The number of blocks is superior to authorized values ' +
                    'for compressed data of ' +
                    'packet {0}: {1} with maximum {2}'.format(
                        self.name,
                        values[self.idx_data_count],
                        self.data_count.maximum,
                    )
                )

            # return the value of the size of compressed data
            return values[self.idx_data_count] * self.data_start.block_size

    return BaseCompressed


def find(iterable, name):
    """
    Find the parameter with name in the list in argument.
    """
    for x in iterable:
        if x.name == name:
            return x

    raise ValueError(
        'Parameter {0} not found in list of parameters'.format(name)
    )


def find_index(iterable, name):
    """
    Find the parameter with name in the list in argument.
    """
    for index, x in enumerate(iterable):
        if x.name == name:
            return index, x

    raise ValueError(
        'Parameter {0} not found in list of parameters'.format(name)
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

try:
    from poppy.core.db.connector import Connector
    from poppy.core.logger import logger
    from poppy.core.tools.exceptions import print_exception, MissingArgument
except:
    print('POPPy framework is not installed!')

try:
    from roc.idb.models.idb import PacketHeader, ItemInfo
    from roc.idb.models.idb import TransferFunction as idb_tf_model
    from roc.idb.exceptions import IdbQueryError
except:
    print('roc.idb plugin is not installed!')

from roc.rpl.packet_parser.packet_cache import PacketCache
from roc.rpl.exceptions import TransferFunctionError


__all__ = [ 'raw_to_eng']

class Enumeration:
    def __init__(self, srdb_id, idb_source, idb_version, model_instance=None):
        self.srdb_id = srdb_id
        self.idb_source = idb_source
        self.idb_version = idb_version
        if model_instance is None:
            self.model_instance = ItemInfo.objects.get(
                srdb_id=srdb_id,
                idb_source=idb_source,
                idb_version=idb_version
            )
        else:
            self.model_instance= model_instance
        transfer_function_values = idb_tf_model.objects.filter(item_info=self.model_instance)

        self.raw_to_eng = {}
        self.eng_to_raw = {}

        for values in transfer_function_values:
            self.raw_to_eng[values.raw] = values.eng
            self.eng_to_raw[values.eng] = values.raw

    @property
    def is_enumeration(self):
        return self.srdb_id.startswith('CIWT')

    def to_raw(self, value):
        return self.eng_to_raw[value]

    def to_eng(self, value):
        return self.raw_to_eng[value]

    @classmethod
    def from_model_instance(cls, model_instance):
        srdb_id = model_instance.srdb_id
        idb_source = model_instance.idb_source
        idb_version = model_instance.idb_version
        return cls(srdb_id, idb_source, idb_version, model_instance=model_instance)

class TransferFunction:

    def __init__(self, srdb_id, idb_source, idb_version, transfer_function=None):
        """
        Initialize instance of TransferFunction class

        :param srdb_id: SRDB ID of the transfer function
        :param idb_source: IDB source
        :param idb_version: IDB version
        :param transfer_function: transfer function data (if provided, skip the query of transfer function data in the database)
        """
        self.srdb_id = srdb_id
        self.idb_source = idb_source
        self.idb_version = idb_version
        try:
            if transfer_function is None:
                # get the IDB from IDBmanager
                self.idb_manager = Connector.manager['MAIN-DB'].idb_manager
                idb = self.idb_manager[(self.idb_source, self.idb_version)]
                # Get transfer function data for current IDB
                self.transfer_function = idb.transfer_function[srdb_id]
            else:
                self.transfer_function = transfer_function
        except:
            raise IdbQueryError(f'Retrieving TF {srdb_id} from IDB {self.idb_version} has failed!')

        self.raw_values, self.eng_values = self.transfer_function_values

        # compute monotonic chunks for reverse interpolation

        # compute the diff between eng values
        eng_values_diff = np.diff(self.eng_values)

        # split the raw and eng arrays to obtain monotonic chunks
        split_index = np.where(np.sign(eng_values_diff[:-1]) * np.sign(eng_values_diff[1:]) < 0 )[0] + 1

        raw_chunks = np.split(self.raw_values,  split_index)
        eng_chunks = np.split(self.eng_values,  split_index)

        # if there is more than one chunk
        chunk_num = len(eng_chunks)
        if chunk_num > 1:
            # ensure the continuity between chunks:
            for chunk_index in range(chunk_num-1):
                raw_chunks[chunk_index] = np.append(raw_chunks[chunk_index], raw_chunks[chunk_index+1][0])
                eng_chunks[chunk_index] = np.append(eng_chunks[chunk_index], eng_chunks[chunk_index+1][0])

        # convert chunk lists to numpy arrays
        eng_chunks = np.array(eng_chunks)
        raw_chunks = np.array(raw_chunks)

        # sort values for each chunk
        sorted_indices = np.argsort(eng_chunks, axis=1)

        self.eng_chunks = np.take_along_axis(eng_chunks, sorted_indices, axis=1)
        self.raw_chunks = np.take_along_axis(raw_chunks, sorted_indices, axis=1)

    @property
    def transfer_function_values(self):
        return zip(*[(float(values[0]), float(values[1])) for values in self.transfer_function])


    @property
    def is_enumeration(self):
        return False

    @staticmethod
    def find_nearest(array, value):
        value = float(value)
        array = np.asarray(array)
        idx = (np.abs(array - value)).argmin()
        return idx

    def to_raw(self, value):

        result_list = []
        # check all monotonic chunks for the reverse interpolation

        for chunk_index, eng_values in enumerate(self.eng_chunks):
            raw_value = np.interp(value, eng_values, self.raw_chunks[chunk_index], left=np.NaN, right=np.NaN)
            if not np.isnan(raw_value):
                # the raw value shall be an integer
                result_list.append(int(raw_value))

        if not result_list:
            raise Exception(f"No raw value found for transfer function '{self.srdb_id}' and eng value '{value}'")
        elif len(result_list) == 1:
            result = result_list[0]
        else:
            # multiple result have been found
            # we need to determine the nearest result
            eng_list = [self.to_eng(raw_result) for raw_result in result_list]
            result = result_list[self.find_nearest(eng_list, value)]
        return result

    def to_eng(self, value):
        return np.interp(value, self.raw_values, self.eng_values, left=np.NaN, right=np.NaN)

class PalisadeTransferFunction(TransferFunction):
    """
    Class used to mitigate differences between MIB and PALISADE
    """

    transfer_function_exceptions = {
        # Discrepancies between TF_CP_BIA_P011 in MIB associated with CIWP0075TC and CIW00107TM
        # and TF_CP_BIA_0011 in PALISADE associated CIWP0040TC and CIW00104TM
        # IMPORTANT : TF_CP_BIA_0011 is obsolete and should not be used (use TF_CP_BIA_P011 instead)
        'CIWP0040TC': ([0, 32767, 32768, 65535], [0, 60, -60, -0.001831]),
        'CIW00104TM': ([0, 32767, 32768, 65535], [0, 60, -60, -0.001831])
    }

    @property
    def transfer_function_values(self):
        if self.srdb_id in self.transfer_function_exceptions:
            logger.warning(f'{self.srdb_id} is not defined in the MIB, use TF at your own risk!')
            return self.transfer_function_exceptions[self.srdb_id]
        else:
            return super().transfer_function_values

    @classmethod
    def from_idb(cls, tf_srdb_id,
                 idb_source='MIB',
                 idb_version=None):

        if not (idb_source and idb_version):
            raise MissingArgument('idb_source and idb_version arguments must be defined to run "raw_to_eng" function!')

        # Check if tf_instance already in the packetCache
        key = (tf_srdb_id, idb_source, idb_version)
        packet_cache = PacketCache()
        tf_instance = packet_cache.transfer_function.get(key, None)

        # If not, get it (and store it in the cache)
        if not tf_instance:
            tf_instance = cls(tf_srdb_id, idb_source, idb_version)
            packet_cache.transfer_function[key] = tf_instance

        return tf_instance


@Connector.if_connected('MAIN-DB')
def raw_to_eng(raw_values, tf_srdb_id,
               idb_source=None,
               idb_version=None):
    """
    Convert input raw values into engineering

    :param raw_values: numpy array containing raw values
    :param tf_srdb_id: string providing the SRDB_ID of the transfer function
    :param idb_source: string with the source of IDB
    :param idb_version: string with the version of IDB
    :return: engineering values
    """

    tf_instance = PalisadeTransferFunction.from_idb(tf_srdb_id,
                                                    idb_source=idb_source,
                                                    idb_version=idb_version)

    # Get instance of TransferFunction for the input transfer function srdb id
    try:
        eng_values = tf_instance.to_eng(raw_values)
    except:
        raise TransferFunctionError(f'Cannot convert "raw" to "eng" for transfer function {tf_srdb_id}!')

    # return engineering values
    return eng_values
